from __future__ import annotations

import functools
from collections import OrderedDict
from enum import Enum
from hashlib import sha256
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Iterator,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from typing_extensions import Final

from dagster import _check as check
from dagster._annotations import deprecated, experimental
from dagster._utils.cached_method import cached_method

if TYPE_CHECKING:
    from dagster._core.definitions.asset_graph import AssetGraph
    from dagster._core.definitions.events import AssetKey
    from dagster._core.events.log import EventLogEntry
    from dagster._core.instance import DagsterInstance


class UnknownValue:
    pass


def foo(x):
    return False


UNKNOWN_VALUE: Final[UnknownValue] = UnknownValue()


class DataVersion(
    NamedTuple(
        "_DataVersion",
        [("value", str)],
    )
):
    """(Experimental) Represents a data version for an asset.

    Args:
        value (str): An arbitrary string representing a data version.
    """

    def __new__(
        cls,
        value: str,
    ):
        return super(DataVersion, cls).__new__(
            cls,
            value=check.str_param(value, "value"),
        )


@experimental
class DataVersionsByPartition(
    NamedTuple(
        "_DataVersionsByPartition", [("data_versions_by_partition", Mapping[str, DataVersion])]
    )
):
    def __new__(
        cls,
        data_versions_by_partition: Mapping[str, Union[str, DataVersion]],
    ):
        check.dict_param(
            data_versions_by_partition,
            "data_versions_by_partition",
            key_type=str,
            value_type=(str, DataVersion),
        )
        return super(DataVersionsByPartition, cls).__new__(
            cls,
            data_versions_by_partition={
                partition: DataVersion(version) if isinstance(version, str) else version
                for partition, version in data_versions_by_partition.items()
            },
        )


DEFAULT_DATA_VERSION: Final[DataVersion] = DataVersion("INITIAL")
NULL_DATA_VERSION: Final[DataVersion] = DataVersion("NULL")
UNKNOWN_DATA_VERSION: Final[DataVersion] = DataVersion("UNKNOWN")


class DataProvenance(
    NamedTuple(
        "_DataProvenance",
        [
            ("code_version", str),
            ("input_data_versions", Mapping["AssetKey", DataVersion]),
            ("is_user_provided", bool),
        ],
    )
):
    """(Experimental) Provenance information for an asset materialization.

    Args:
        code_version (str): The code version of the op that generated a materialization.
        input_data_versions (Mapping[AssetKey, DataVersion]): The data versions of the
            inputs used to generate a materialization.
        is_user_provided (bool): True if the data version was provided directly by the user, false
            if it was generated by Dagster.
    """

    def __new__(
        cls,
        code_version: str,
        input_data_versions: Mapping["AssetKey", DataVersion],
        is_user_provided: bool,
    ):
        from dagster._core.definitions.events import AssetKey

        return super(DataProvenance, cls).__new__(
            cls,
            code_version=check.str_param(code_version, "code_version"),
            input_data_versions=check.mapping_param(
                input_data_versions,
                "input_data_versions",
                key_type=AssetKey,
                value_type=DataVersion,
            ),
            is_user_provided=check.bool_param(is_user_provided, "is_user_provided"),
        )

    @staticmethod
    def from_tags(tags: Mapping[str, str]) -> Optional[DataProvenance]:
        from dagster._core.definitions.events import AssetKey

        code_version = tags.get(CODE_VERSION_TAG)
        if code_version is None:
            return None
        input_data_versions = {
            # Everything after the 2nd slash is the asset key
            AssetKey.from_user_string(k.split("/", maxsplit=2)[-1]): DataVersion(tags[k])
            for k, v in tags.items()
            if k.startswith(INPUT_DATA_VERSION_TAG_PREFIX)
            or k.startswith(_OLD_INPUT_DATA_VERSION_TAG_PREFIX)
        }
        is_user_provided = tags.get(DATA_VERSION_IS_USER_PROVIDED_TAG) == "true"
        return DataProvenance(code_version, input_data_versions, is_user_provided)

    @property
    @deprecated
    def input_logical_versions(self) -> Mapping["AssetKey", DataVersion]:
        return self.input_data_versions


# ########################
# ##### TAG KEYS
# ########################

DATA_VERSION_TAG: Final[str] = "dagster/data_version"
_OLD_DATA_VERSION_TAG: Final[str] = "dagster/logical_version"
CODE_VERSION_TAG: Final[str] = "dagster/code_version"
INPUT_DATA_VERSION_TAG_PREFIX: Final[str] = "dagster/input_data_version"
_OLD_INPUT_DATA_VERSION_TAG_PREFIX: Final[str] = "dagster/input_logical_version"
INPUT_EVENT_POINTER_TAG_PREFIX: Final[str] = "dagster/input_event_pointer"
DATA_VERSION_IS_USER_PROVIDED_TAG = "dagster/data_version_is_user_provided"


def read_input_data_version_from_tags(
    tags: Mapping[str, str], input_key: "AssetKey"
) -> Optional[DataVersion]:
    value = tags.get(
        get_input_data_version_tag(input_key, prefix=INPUT_DATA_VERSION_TAG_PREFIX)
    ) or tags.get(get_input_data_version_tag(input_key, prefix=_OLD_INPUT_DATA_VERSION_TAG_PREFIX))
    return DataVersion(value) if value is not None else None


def get_input_data_version_tag(
    input_key: "AssetKey", prefix: str = INPUT_DATA_VERSION_TAG_PREFIX
) -> str:
    return f"{prefix}/{input_key.to_user_string()}"


def get_input_event_pointer_tag(input_key: "AssetKey") -> str:
    return f"{INPUT_EVENT_POINTER_TAG_PREFIX}/{input_key.to_user_string()}"


# ########################
# ##### COMPUTE / EXTRACT
# ########################


def compute_logical_data_version(
    code_version: Union[str, UnknownValue],
    input_data_versions: Mapping["AssetKey", DataVersion],
) -> DataVersion:
    """Compute a data version for a value as a hash of input data versions and code version.

    Args:
        code_version (str): The code version of the computation.
        input_data_versions (Mapping[AssetKey, DataVersion]): The data versions of the inputs.

    Returns:
        DataVersion: The computed version as a `DataVersion`.
    """
    from dagster._core.definitions.events import AssetKey

    check.inst_param(code_version, "code_version", (str, UnknownValue))
    check.mapping_param(
        input_data_versions, "input_versions", key_type=AssetKey, value_type=DataVersion
    )

    if (
        isinstance(code_version, UnknownValue)
        or UNKNOWN_DATA_VERSION in input_data_versions.values()
    ):
        return UNKNOWN_DATA_VERSION

    ordered_input_versions = [
        input_data_versions[k] for k in sorted(input_data_versions.keys(), key=str)
    ]
    all_inputs = (code_version, *(v.value for v in ordered_input_versions))

    hash_sig = sha256()
    hash_sig.update(bytearray("".join(all_inputs), "utf8"))
    return DataVersion(hash_sig.hexdigest())


def extract_data_version_from_entry(
    entry: EventLogEntry,
) -> Optional[DataVersion]:
    tags = entry.tags or {}
    value = tags.get(DATA_VERSION_TAG) or tags.get(_OLD_DATA_VERSION_TAG)
    return None if value is None else DataVersion(value)


def extract_data_provenance_from_entry(
    entry: EventLogEntry,
) -> Optional[DataProvenance]:
    tags = entry.tags or {}
    return DataProvenance.from_tags(tags)


# ########################
# ##### STALENESS OPERATIONS
# ########################


class StaleStatus(Enum):
    MISSING = "MISSING"
    STALE = "STALE"
    FRESH = "FRESH"


@functools.total_ordering
class StaleCauseCategory(Enum):
    CODE = "CODE"
    DATA = "DATA"
    DEPENDENCIES = "DEPENDENCIES"

    def __lt__(self, other: "StaleCauseCategory"):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class StaleCause(NamedTuple):
    key: AssetKey
    category: StaleCauseCategory
    reason: str
    dependency: Optional[AssetKey] = None
    children: Optional[Sequence["StaleCause"]] = None


class CachingStaleStatusResolver:
    """Used to resolve data version information. Avoids redundant database
    calls that would otherwise occur. Intended for use within the scope of a
    single "request" (e.g. GQL request, RunRequest resolution).
    """

    _instance: "DagsterInstance"
    _asset_graph: Optional["AssetGraph"]
    _asset_graph_load_fn: Optional[Callable[[], "AssetGraph"]]

    def __init__(
        self,
        instance: "DagsterInstance",
        asset_graph: Union["AssetGraph", Callable[[], "AssetGraph"]],
    ):
        from dagster._core.definitions.asset_graph import AssetGraph

        self._instance = instance
        if isinstance(asset_graph, AssetGraph):
            self._asset_graph = asset_graph
            self._asset_graph_load_fn = None
        else:
            self._asset_graph = None
            self._asset_graph_load_fn = asset_graph

    def get_status(self, key: AssetKey) -> StaleStatus:
        return self._get_status(key=key)

    def get_stale_causes(self, key: AssetKey) -> Sequence[StaleCause]:
        return self._get_stale_causes(key=key)

    def get_stale_root_causes(self, key: AssetKey) -> Sequence[StaleCause]:
        return self._get_stale_root_causes(key=key)

    def get_current_data_version(self, key: AssetKey) -> DataVersion:
        return self._get_current_data_version(key=key)

    @cached_method
    def _get_status(self, key: AssetKey) -> StaleStatus:
        current_version = self._get_current_data_version(key=key)
        if current_version == NULL_DATA_VERSION:
            return StaleStatus.MISSING
        elif self.asset_graph.is_source(key) or self._is_partitioned_or_downstream(key=key):
            return StaleStatus.FRESH
        else:
            causes = self._get_stale_causes(key=key)
            return StaleStatus.FRESH if len(causes) == 0 else StaleStatus.STALE

    @cached_method
    def _get_stale_causes(self, key: AssetKey) -> Sequence[StaleCause]:
        current_version = self._get_current_data_version(key=key)
        if (
            current_version == NULL_DATA_VERSION
            or self.asset_graph.is_source(key)
            or self._is_partitioned_or_downstream(key=key)
        ):
            return []
        else:
            return list(self._get_stale_causes_materialized(key))

    def _get_stale_causes_materialized(self, key: AssetKey) -> Iterator[StaleCause]:
        code_version = self.asset_graph.get_code_version(key)
        provenance = self._get_current_data_provenance(key=key)
        dependency_keys = self.asset_graph.get_parents(key)

        # only used if no provenance available
        materialization = check.not_none(self._get_latest_materialization_event(key=key))
        materialization_time = materialization.timestamp

        if provenance:
            if code_version and code_version != provenance.code_version:
                yield StaleCause(key, StaleCauseCategory.CODE, "has a new code version")

            removed_deps = set(provenance.input_data_versions.keys()) - set(dependency_keys)
            for dep_key in removed_deps:
                yield StaleCause(
                    key,
                    StaleCauseCategory.DEPENDENCIES,
                    f"removed dependency on {dep_key.to_user_string()}",
                    dep_key,
                )

        for dep_key in sorted(dependency_keys):
            if self._get_status(key=dep_key) == StaleStatus.STALE:
                yield StaleCause(
                    key,
                    StaleCauseCategory.DATA,
                    "stale dependency",
                    dep_key,
                    self._get_stale_causes(key=dep_key),
                )
            elif provenance:
                if dep_key not in provenance.input_data_versions:
                    yield StaleCause(
                        key,
                        StaleCauseCategory.DEPENDENCIES,
                        f"has a new dependency on {dep_key.to_user_string()}",
                        dep_key,
                    )
                elif provenance.input_data_versions[dep_key] != self._get_current_data_version(
                    key=dep_key
                ):
                    report_data_version = self.asset_graph.get_code_version(
                        dep_key
                    ) is not None or self._is_current_data_version_user_provided(key=dep_key)
                    yield StaleCause(
                        key,
                        StaleCauseCategory.DATA,
                        "has a new dependency data version"
                        if report_data_version
                        else "has a new dependency materialization",
                        dep_key,
                        [
                            StaleCause(
                                dep_key,
                                StaleCauseCategory.DATA,
                                # Assets with no defined code version will get a new data version on
                                # every materialization, so we just report new materialization for
                                # this case since the user likely hasn't engaged with data versions.
                                "has a new data version"
                                if report_data_version
                                else "has a new materialization",
                            )
                        ],
                    )
            # If no provenance and dep is a materializable asset, then use materialization
            # timestamps instead of versions this should be removable eventually since
            # provenance is on all newer materializations. If dep is a source, then we'll never
            # provide a stale reason here.
            elif not self.asset_graph.is_source(dep_key):
                dep_materialization = self._get_latest_materialization_event(key=dep_key)
                if dep_materialization is None:
                    # The input must be new if it has no materialization
                    yield StaleCause(key, StaleCauseCategory.DATA, "has a new input", dep_key)
                elif dep_materialization.timestamp > materialization_time:
                    yield StaleCause(
                        key,
                        StaleCauseCategory.DATA,
                        "has a new dependency materialization",
                        dep_key,
                        [
                            StaleCause(
                                dep_key,
                                StaleCauseCategory.DATA,
                                "has a new materialization",
                            )
                        ],
                    )

    @cached_method
    def _get_stale_root_causes(self, key: AssetKey) -> Sequence[StaleCause]:
        causes = self._get_stale_causes(key=key)
        root_pairs = sorted([pair for cause in causes for pair in self._gather_leaves(cause)])
        # After sorting the pairs, we can drop the level and de-dup using an
        # ordered dict as an ordered set. This will give us unique root causes,
        # sorted by level.
        roots: Dict[StaleCause, None] = OrderedDict()
        for root_cause in [leaf_cause for _, leaf_cause in root_pairs]:
            roots[root_cause] = None
        return list(roots.keys())

    # The leaves of the cause tree for an asset are the root causes of its staleness.
    def _gather_leaves(self, cause: StaleCause, level: int = 0) -> Iterator[Tuple[int, StaleCause]]:
        if cause.children is None:
            yield (level, cause)
        else:
            for child in cause.children:
                yield from self._gather_leaves(child, level=level + 1)

    @property
    def asset_graph(self) -> "AssetGraph":
        if self._asset_graph is None:
            self._asset_graph = check.not_none(self._asset_graph_load_fn)()
        return self._asset_graph

    @cached_method
    def _get_current_data_version(self, *, key: AssetKey) -> DataVersion:
        is_source = self.asset_graph.is_source(key)
        event = self._instance.get_latest_data_version_record(
            key,
            is_source,
        )
        if event is None and is_source:
            return DEFAULT_DATA_VERSION
        elif event is None:
            return NULL_DATA_VERSION
        else:
            data_version = extract_data_version_from_entry(event.event_log_entry)
            return data_version or DEFAULT_DATA_VERSION

    @cached_method
    def _get_latest_materialization_event(self, *, key: AssetKey) -> Optional[EventLogEntry]:
        return self._instance.get_latest_materialization_event(key)

    @cached_method
    def _is_current_data_version_user_provided(self, *, key: AssetKey) -> bool:
        if self.asset_graph.is_source(key):
            return True
        else:
            provenance = self._get_current_data_provenance(key=key)
            return provenance is not None and provenance.is_user_provided

    @cached_method
    def _get_current_data_provenance(self, *, key: AssetKey) -> Optional[DataProvenance]:
        materialization = self._get_latest_materialization_event(key=key)
        if materialization is None:
            return None
        else:
            return extract_data_provenance_from_entry(materialization)

    @cached_method
    def _is_partitioned_or_downstream(self, *, key: AssetKey) -> bool:
        if self.asset_graph.get_partitions_def(key):
            return True
        elif self.asset_graph.is_source(key):
            return False
        else:
            return any(
                self._is_partitioned_or_downstream(key=dep_key)
                for dep_key in self.asset_graph.get_parents(key)
            )

    # Volatility means that an asset is assumed to be constantly changing. We assume that observable
    # source assets are non-volatile, since the primary purpose of the observation function is to
    # determine if a source asset has changed. We assume that regular assets are volatile if they
    # are at the root of the graph (have no dependencies) or are downstream of a volatile asset.
    @cached_method
    def _is_volatile(self, *, key: AssetKey) -> bool:
        if self.asset_graph.is_source(key):
            return self.asset_graph.is_observable(key)
        else:
            deps = self.asset_graph.get_parents(key)
            return len(deps) == 0 or any(self._is_volatile(key=dep_key) for dep_key in deps)



from dagster import In, OpDefinition, Out, Output, job


def test_op_def_direct():
    def the_op_fn(_, inputs):
        assert inputs["x"] == 5
        yield Output(inputs["x"] + 1, output_name="the_output")

    op_def = OpDefinition(
        the_op_fn, "the_op", ins={"x": In(dagster_type=int)}, outs={"the_output": Out(int)}
    )

    @job
    def the_job(x):
        op_def(x)

    result = the_job.execute_in_process(input_values={"x": 5})
    assert result.success



from .base import (
    AssetRecord as AssetRecord,
    EventLogRecord as EventLogRecord,
    EventLogStorage as EventLogStorage,
)
from .in_memory import InMemoryEventLogStorage as InMemoryEventLogStorage
from .polling_event_watcher import SqlPollingEventWatcher as SqlPollingEventWatcher
from .schema import (
    AssetKeyTable as AssetKeyTable,
    DynamicPartitionsTable as DynamicPartitionsTable,
    SqlEventLogStorageMetadata as SqlEventLogStorageMetadata,
    SqlEventLogStorageTable as SqlEventLogStorageTable,
)
from .sql_event_log import SqlEventLogStorage as SqlEventLogStorage
from .sqlite import (
    ConsolidatedSqliteEventLogStorage as ConsolidatedSqliteEventLogStorage,
    SqliteEventLogStorage as SqliteEventLogStorage,
)



from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Union

from dagster import (
    InputContext,
    MetadataValue,
    MultiPartitionKey,
    OutputContext,
    _check as check,
)
from dagster._core.storage.memoizable_io_manager import MemoizableIOManager

if TYPE_CHECKING:
    from upath import UPath


class UPathIOManager(MemoizableIOManager):
    """Abstract IOManager base class compatible with local and cloud storage via `universal-pathlib` and `fsspec`.

    Features:
     - handles partitioned assets
     - handles loading a single upstream partition
     - handles loading multiple upstream partitions (with respect to <PyObject object="PartitionMapping" />)
     - the `get_metadata` method can be customized to add additional metadata to the output
     - the `allow_missing_partitions` metadata value can be set to `True` to skip missing partitions
       (the default behavior is to raise an error)

    """

    extension: Optional[str] = None  # override in child class

    def __init__(
        self,
        base_path: Optional["UPath"] = None,
    ):
        from upath import UPath

        assert not self.extension or "." in self.extension
        self._base_path = base_path or UPath(".")

    @abstractmethod
    def dump_to_path(self, context: OutputContext, obj: Any, path: "UPath"):
        """Child classes should override this method to write the object to the filesystem."""

    @abstractmethod
    def load_from_path(self, context: InputContext, path: "UPath") -> Any:
        """Child classes should override this method to load the object from the filesystem."""

    def get_metadata(
        self,
        context: OutputContext,
        obj: Any,
    ) -> Dict[str, MetadataValue]:
        """Child classes should override this method to add custom metadata to the outputs."""
        return {}

    # Read/write operations on paths can generally be handled by methods on the
    # UPath class, but when the backend requires credentials, this isn't
    # always possible. Override these path_* methods to provide custom
    # implementations for targeting backends that require authentication.

    def unlink(self, path: "UPath") -> None:
        """Remove the file or object at the provided path."""
        path.unlink()

    def path_exists(self, path: "UPath") -> bool:
        """Check if a file or object exists at the provided path."""
        return path.exists()

    def make_directory(self, path: "UPath"):
        """Create a directory at the provided path.

        Override as a no-op if the target backend doesn't use directories.
        """
        path.mkdir(parents=True, exist_ok=True)

    def has_output(self, context: OutputContext) -> bool:
        return self.path_exists(self._get_path(context))

    def _with_extension(self, path: "UPath") -> "UPath":
        from upath import UPath

        return UPath(f"{path}{self.extension}") if self.extension else path

    def _get_path_without_extension(self, context: Union[InputContext, OutputContext]) -> "UPath":
        if context.has_asset_key:
            context_path = self.get_asset_relative_path(context)
        else:
            # we are dealing with an op output
            context_path = self.get_op_output_relative_path(context)

        return self._base_path.joinpath(context_path)

    def get_asset_relative_path(self, context: Union[InputContext, OutputContext]) -> "UPath":
        from upath import UPath

        # we are not using context.get_asset_identifier() because it already includes the partition_key
        return UPath(*context.asset_key.path)

    def get_op_output_relative_path(self, context: Union[InputContext, OutputContext]) -> "UPath":
        from upath import UPath

        return UPath(*context.get_identifier())

    def get_loading_input_log_message(self, path: "UPath") -> str:
        return f"Loading file from: {path}"

    def get_writing_output_log_message(self, path: "UPath") -> str:
        return f"Writing file at: {path}"

    def _get_path(self, context: Union[InputContext, OutputContext]) -> "UPath":
        """Returns the I/O path for a given context.
        Should not be used with partitions (use `_get_paths_for_partitions` instead).
        """
        path = self._get_path_without_extension(context)
        return self._with_extension(path)

    def _get_paths_for_partitions(
        self, context: Union[InputContext, OutputContext]
    ) -> Dict[str, "UPath"]:
        """Returns a dict of partition_keys into I/O paths for a given context."""
        if not context.has_asset_partitions:
            raise TypeError(
                f"Detected {context.dagster_type.typing_type} input type "
                "but the asset is not partitioned"
            )

        def _formatted_multipartitioned_path(partition_key: MultiPartitionKey) -> str:
            ordered_dimension_keys = [
                key[1]
                for key in sorted(partition_key.keys_by_dimension.items(), key=lambda x: x[0])
            ]
            return "/".join(ordered_dimension_keys)

        formatted_partition_keys = [
            _formatted_multipartitioned_path(pk) if isinstance(pk, MultiPartitionKey) else pk
            for pk in context.asset_partition_keys
        ]

        asset_path = self._get_path_without_extension(context)
        return {
            partition_key: self._with_extension(asset_path / partition_key)
            for partition_key in formatted_partition_keys
        }

    def _get_multipartition_backcompat_paths(
        self, context: Union[InputContext, OutputContext]
    ) -> Mapping[str, "UPath"]:
        if not context.has_asset_partitions:
            raise TypeError(
                f"Detected {context.dagster_type.typing_type} input type "
                "but the asset is not partitioned"
            )

        partition_keys = context.asset_partition_keys

        asset_path = self._get_path_without_extension(context)
        return {
            partition_key: self._with_extension(asset_path / partition_key)
            for partition_key in partition_keys
            if isinstance(partition_key, MultiPartitionKey)
        }

    def _load_single_input(
        self, path: "UPath", context: InputContext, backcompat_path: Optional["UPath"] = None
    ) -> Any:
        context.log.debug(self.get_loading_input_log_message(path))
        try:
            obj = self.load_from_path(context=context, path=path)
        except FileNotFoundError as e:
            if backcompat_path is not None:
                try:
                    obj = self.load_from_path(context=context, path=backcompat_path)
                    context.log.debug(
                        f"File not found at {path}. Loaded instead from backcompat path:"
                        f" {backcompat_path}"
                    )
                except FileNotFoundError:
                    raise e
            else:
                raise e

        context.add_input_metadata({"path": MetadataValue.path(str(path))})
        return obj

    def _load_multiple_inputs(self, context: InputContext) -> Dict[str, Any]:
        # load multiple partitions
        allow_missing_partitions = (
            context.metadata.get("allow_missing_partitions", False)
            if context.metadata is not None
            else False
        )

        objs: Dict[str, Any] = {}
        paths = self._get_paths_for_partitions(context)
        backcompat_paths = self._get_multipartition_backcompat_paths(context)

        context.log.debug(f"Loading {len(paths)} partitions...")

        for partition_key, path in paths.items():
            context.log.debug(f"Loading partition from {path} using {self.__class__.__name__}")
            try:
                obj = self.load_from_path(context=context, path=path)
                objs[partition_key] = obj
            except FileNotFoundError as e:
                backcompat_path = backcompat_paths.get(partition_key)
                if backcompat_path is not None:
                    obj = self.load_from_path(context=context, path=backcompat_path)
                    objs[partition_key] = obj

                if not allow_missing_partitions and objs.get(partition_key) is None:
                    raise e

                context.log.debug(
                    f"Couldn't load partition {path} and skipped it "
                    "because the input metadata includes allow_missing_partitions=True"
                )

        # TODO: context.add_output_metadata fails in the partitioned context. this should be fixed?
        return objs

    def load_input(self, context: InputContext) -> Union[Any, Dict[str, Any]]:
        # If no asset key, we are dealing with an op output which is always non-partitioned
        if not context.has_asset_key or not context.has_asset_partitions:
            path = self._get_path(context)
            return self._load_single_input(path, context)
        else:
            asset_partition_keys = context.asset_partition_keys
            if len(asset_partition_keys) == 0:
                return None
            elif len(asset_partition_keys) == 1:
                paths = self._get_paths_for_partitions(context)
                check.invariant(len(paths) == 1, f"Expected 1 path, but got {len(paths)}")
                path = list(paths.values())[0]
                backcompat_paths = self._get_multipartition_backcompat_paths(context)
                backcompat_path = (
                    None if not backcompat_paths else list(backcompat_paths.values())[0]
                )

                return self._load_single_input(path, context, backcompat_path)
            else:  # we are dealing with multiple partitions of an asset
                type_annotation = context.dagster_type.typing_type
                if type_annotation != Any and not is_dict_type(type_annotation):
                    check.failed(
                        "Loading an input that corresponds to multiple partitions, but the"
                        " type annotation on the op input is not a dict, Dict, Mapping, or"
                        f" Any: is '{type_annotation}'."
                    )

                return self._load_multiple_inputs(context)

    def handle_output(self, context: OutputContext, obj: Any):
        if context.dagster_type.typing_type == type(None):
            check.invariant(
                obj is None,
                (
                    "Output had Nothing type or 'None' annotation, but handle_output received"
                    f" value that was not None and was of type {type(obj)}."
                ),
            )
            return None

        if context.has_asset_partitions:
            paths = self._get_paths_for_partitions(context)
            assert len(paths) == 1
            path = list(paths.values())[0]
        else:
            path = self._get_path(context)
        self.make_directory(path.parent)
        context.log.debug(self.get_writing_output_log_message(path))
        self.dump_to_path(context=context, obj=obj, path=path)

        metadata = {"path": MetadataValue.path(str(path))}
        custom_metadata = self.get_metadata(context=context, obj=obj)
        metadata.update(custom_metadata)  # type: ignore

        context.add_output_metadata(metadata)


def is_dict_type(type_obj) -> bool:
    if type_obj == dict:
        return True

    if hasattr(type_obj, "__origin__") and type_obj.__origin__ in (dict, Dict, Mapping):
        return True

    return False



from typing import Optional

import pendulum

import dagster._check as check
from dagster._core.definitions.asset_reconciliation_sensor import (
    AssetReconciliationCursor,
    reconcile,
)
from dagster._core.definitions.external_asset_graph import ExternalAssetGraph
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.instance import DagsterInstance
from dagster._core.storage.dagster_run import DagsterRunStatus
from dagster._core.storage.tags import AUTO_MATERIALIZE_TAG
from dagster._core.workspace.context import IWorkspaceProcessContext
from dagster._daemon.daemon import DaemonIterator, IntervalDaemon

CURSOR_KEY = "ASSET_DAEMON_CURSOR"
ASSET_DAEMON_PAUSED_KEY = "ASSET_DAEMON_PAUSED"

EVALUATIONS_TTL_DAYS = 7


def get_auto_materialize_paused(instance: DagsterInstance) -> bool:
    return (
        instance.daemon_cursor_storage.get_cursor_values({ASSET_DAEMON_PAUSED_KEY}).get(
            ASSET_DAEMON_PAUSED_KEY
        )
        != "false"
    )


def set_auto_materialize_paused(instance: DagsterInstance, paused: bool):
    instance.daemon_cursor_storage.set_cursor_values(
        {ASSET_DAEMON_PAUSED_KEY: "true" if paused else "false"}
    )


def _get_raw_cursor(instance: DagsterInstance) -> Optional[str]:
    return instance.daemon_cursor_storage.get_cursor_values({CURSOR_KEY}).get(CURSOR_KEY)


def get_current_evaluation_id(instance: DagsterInstance) -> Optional[int]:
    raw_cursor = _get_raw_cursor(instance)
    return (
        AssetReconciliationCursor.get_evaluation_id_from_serialized(raw_cursor)
        if raw_cursor
        else None
    )


class AssetDaemon(IntervalDaemon):
    def __init__(self, interval_seconds: int):
        super().__init__(interval_seconds=interval_seconds)

    @classmethod
    def daemon_type(cls) -> str:
        return "ASSET"

    def run_iteration(
        self,
        workspace_process_context: IWorkspaceProcessContext,
    ) -> DaemonIterator:
        instance = workspace_process_context.instance

        if get_auto_materialize_paused(instance):
            yield
            return

        schedule_storage = check.not_none(
            instance.schedule_storage,
            "Auto materialization requires schedule storage to be configured",
        )

        if not schedule_storage.supports_auto_materialize_asset_evaluations:
            self._logger.warning(
                "Auto materialize evaluations are not getting logged. Run `dagster instance"
                " migrate` to enable."
            )

        workspace = workspace_process_context.create_request_context()
        asset_graph = ExternalAssetGraph.from_workspace(workspace)
        target_asset_keys = {
            target_key
            for target_key in asset_graph.non_source_asset_keys
            if asset_graph.get_auto_materialize_policy(target_key) is not None
        }

        if not target_asset_keys:
            yield
            return

        raw_cursor = _get_raw_cursor(instance)
        cursor = (
            AssetReconciliationCursor.from_serialized(raw_cursor, asset_graph)
            if raw_cursor
            else AssetReconciliationCursor.empty()
        )

        run_requests, new_cursor, evaluations = reconcile(
            asset_graph=asset_graph,
            target_asset_keys=target_asset_keys,
            instance=instance,
            cursor=cursor,
            run_tags=None,
        )

        for run_request in run_requests:
            yield

            asset_keys = check.not_none(run_request.asset_selection)
            check.invariant(len(asset_keys) > 0)

            repo_handle = asset_graph.get_repository_handle(asset_keys[0])

            # Check that all asset keys are from the same repo
            for key in asset_keys[1:]:
                check.invariant(repo_handle == asset_graph.get_repository_handle(key))

            location_name = repo_handle.code_location_origin.location_name
            repository_name = repo_handle.repository_name
            job_name = check.not_none(asset_graph.get_implicit_job_name_for_assets(asset_keys))

            code_location = workspace.get_code_location(location_name)
            external_job = code_location.get_external_job(
                JobSubsetSelector(
                    location_name=location_name,
                    repository_name=repository_name,
                    job_name=job_name,
                    op_selection=None,
                    asset_selection=asset_keys,
                )
            )

            tags = {
                **run_request.tags,
                AUTO_MATERIALIZE_TAG: "true",
                **instance.auto_materialize_run_tags,
            }

            external_execution_plan = code_location.get_external_execution_plan(
                external_job,
                run_request.run_config,
                step_keys_to_execute=None,
                known_state=None,
                instance=instance,
            )
            execution_plan_snapshot = external_execution_plan.execution_plan_snapshot

            run = instance.create_run(
                job_name=external_job.name,
                run_id=None,
                run_config=None,
                resolved_op_selection=None,
                step_keys_to_execute=None,
                status=DagsterRunStatus.NOT_STARTED,
                op_selection=None,
                root_run_id=None,
                parent_run_id=None,
                tags=tags,
                job_snapshot=external_job.job_snapshot,
                execution_plan_snapshot=execution_plan_snapshot,
                parent_job_snapshot=external_job.parent_job_snapshot,
                external_job_origin=external_job.get_external_origin(),
                job_code_origin=external_job.get_python_origin(),
                asset_selection=frozenset(asset_keys),
            )
            instance.submit_run(run.run_id, workspace)

        instance.daemon_cursor_storage.set_cursor_values({CURSOR_KEY: new_cursor.serialize()})

        # We enforce uniqueness per (asset key, evaluation id). Store the evaluations after the cursor,
        # so that if the daemon crashes and doesn't update the cursor we don't try to write duplicates.
        if schedule_storage.supports_auto_materialize_asset_evaluations:
            schedule_storage.add_auto_materialize_asset_evaluations(
                new_cursor.evaluation_id, evaluations
            )
            schedule_storage.purge_asset_evaluations(
                before=pendulum.now("UTC").subtract(days=EVALUATIONS_TTL_DAYS).timestamp(),
            )



from .base import (
    RunCoordinator as RunCoordinator,
    SubmitRunContext as SubmitRunContext,
)
from .default_run_coordinator import DefaultRunCoordinator as DefaultRunCoordinator
from .queued_run_coordinator import QueuedRunCoordinator as QueuedRunCoordinator



import datetime
import logging
import sys
import threading
import time
import uuid
from contextlib import AbstractContextManager, ExitStack, contextmanager
from types import TracebackType
from typing import Callable, Dict, Iterable, Iterator, Mapping, Optional, Sequence, Type

import pendulum
from typing_extensions import Self

import dagster._check as check
from dagster._core.host_representation.grpc_server_registry import GrpcServerRegistry
from dagster._core.instance import DagsterInstance
from dagster._core.workspace.context import IWorkspaceProcessContext, WorkspaceProcessContext
from dagster._core.workspace.load_target import WorkspaceLoadTarget
from dagster._daemon.asset_daemon import AssetDaemon
from dagster._daemon.auto_run_reexecution.event_log_consumer import EventLogConsumerDaemon
from dagster._daemon.daemon import (
    BackfillDaemon,
    DagsterDaemon,
    MonitoringDaemon,
    SchedulerDaemon,
    SensorDaemon,
)
from dagster._daemon.run_coordinator.queued_run_coordinator_daemon import QueuedRunCoordinatorDaemon
from dagster._daemon.types import DaemonHeartbeat, DaemonStatus
from dagster._utils.interrupts import raise_interrupts_as
from dagster._utils.log import configure_loggers

# How long beyond the expected heartbeat will the daemon be considered healthy
DEFAULT_DAEMON_HEARTBEAT_TOLERANCE_SECONDS = 300

# Default interval at which daemons run
DEFAULT_DAEMON_INTERVAL_SECONDS = 30

# Interval at which heartbeats are posted
DEFAULT_HEARTBEAT_INTERVAL_SECONDS = 30

# How long after an error is raised in a daemon that it's still included in the status for that daemon
DEFAULT_DAEMON_ERROR_INTERVAL_SECONDS = 300

THREAD_CHECK_INTERVAL = 5

HEARTBEAT_CHECK_INTERVAL = 15

RELOAD_WORKSPACE_INTERVAL = 60

DAEMON_GRPC_SERVER_RELOAD_INTERVAL = 60
DAEMON_GRPC_SERVER_HEARTBEAT_TTL = 120


def _sorted_quoted(strings: Iterable[str]) -> str:
    return "[" + ", ".join([f"'{s}'" for s in sorted(list(strings))]) + "]"


def create_daemons_from_instance(instance: DagsterInstance) -> Sequence[DagsterDaemon]:
    return [
        create_daemon_of_type(daemon_type, instance)
        for daemon_type in instance.get_required_daemon_types()
    ]


def create_daemon_grpc_server_registry(
    instance: DagsterInstance, code_server_log_level: str = "INFO"
) -> GrpcServerRegistry:
    return GrpcServerRegistry(
        instance_ref=instance.get_ref(),
        reload_interval=0,  # refresh_workspace call handles the reload
        heartbeat_ttl=DAEMON_GRPC_SERVER_HEARTBEAT_TTL,
        startup_timeout=instance.code_server_process_startup_timeout,
        log_level=code_server_log_level,
        wait_for_processes_on_shutdown=instance.wait_for_local_code_server_processes_on_shutdown,
    )


@contextmanager
def daemon_controller_from_instance(
    instance: DagsterInstance,
    workspace_load_target: WorkspaceLoadTarget,
    heartbeat_interval_seconds: int = DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
    heartbeat_tolerance_seconds: int = DEFAULT_DAEMON_HEARTBEAT_TOLERANCE_SECONDS,
    gen_daemons: Callable[
        [DagsterInstance], Iterable[DagsterDaemon]
    ] = create_daemons_from_instance,
    error_interval_seconds: int = DEFAULT_DAEMON_ERROR_INTERVAL_SECONDS,
    code_server_log_level: str = "info",
) -> Iterator["DagsterDaemonController"]:
    check.inst_param(instance, "instance", DagsterInstance)
    check.inst_param(workspace_load_target, "workspace_load_target", WorkspaceLoadTarget)
    grpc_server_registry: Optional[GrpcServerRegistry] = None
    with ExitStack() as stack:
        grpc_server_registry = stack.enter_context(
            create_daemon_grpc_server_registry(instance, code_server_log_level)
        )
        daemons = [stack.enter_context(daemon) for daemon in gen_daemons(instance)]
        workspace_process_context = stack.enter_context(
            WorkspaceProcessContext(
                instance=instance,
                workspace_load_target=workspace_load_target,
                grpc_server_registry=grpc_server_registry,
            )
        )
        controller = stack.enter_context(
            DagsterDaemonController(
                workspace_process_context,
                daemons,
                heartbeat_interval_seconds=heartbeat_interval_seconds,
                heartbeat_tolerance_seconds=heartbeat_tolerance_seconds,
                error_interval_seconds=error_interval_seconds,
                grpc_server_registry=grpc_server_registry,
            )
        )

        yield controller


class DagsterDaemonController(AbstractContextManager):
    _daemon_uuid: str
    _daemons: Dict[str, DagsterDaemon]
    _grpc_server_registry: Optional[GrpcServerRegistry]
    _daemon_threads: Dict[str, threading.Thread]
    _workspace_process_context: IWorkspaceProcessContext
    _instance: DagsterInstance
    _heartbeat_interval_seconds: float
    _heartbeat_tolerance_seconds: float
    _daemon_shutdown_event: threading.Event
    _logger: logging.Logger
    _last_healthy_heartbeat_times: Dict[str, float]
    _start_time: datetime.datetime

    def __init__(
        self,
        workspace_process_context: IWorkspaceProcessContext,
        daemons: Sequence[DagsterDaemon],
        grpc_server_registry: Optional[GrpcServerRegistry] = None,
        heartbeat_interval_seconds: float = DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
        heartbeat_tolerance_seconds: float = DEFAULT_DAEMON_HEARTBEAT_TOLERANCE_SECONDS,
        error_interval_seconds: int = DEFAULT_DAEMON_ERROR_INTERVAL_SECONDS,
        handler: str = "default",
    ):
        self._daemon_uuid = str(uuid.uuid4())

        self._daemons = {}
        self._daemon_threads = {}

        self._workspace_process_context = workspace_process_context
        self._instance = workspace_process_context.instance
        self._daemons = {daemon.daemon_type(): daemon for daemon in daemons}

        self._heartbeat_interval_seconds = check.numeric_param(
            heartbeat_interval_seconds, "heartbeat_interval_seconds"
        )

        self._heartbeat_tolerance_seconds = check.numeric_param(
            heartbeat_tolerance_seconds, "heartbeat_tolerance_seconds"
        )

        self._grpc_server_registry = grpc_server_registry

        if not self._daemons:
            raise Exception("No daemons configured on the DagsterInstance")

        self._daemon_shutdown_event = threading.Event()

        configure_loggers(handler=handler)

        self._logger = logging.getLogger("dagster.daemon")
        self._logger.info(
            "Instance is configured with the following daemons: %s",
            _sorted_quoted(type(daemon).__name__ for daemon in self.daemons),
        )

        self._last_healthy_heartbeat_times = {}

        for daemon_type, daemon in self._daemons.items():
            self._daemon_threads[daemon_type] = threading.Thread(
                target=daemon.run_daemon_loop,
                args=(
                    workspace_process_context,
                    self._daemon_uuid,
                    self._daemon_shutdown_event,
                    heartbeat_interval_seconds,
                    error_interval_seconds,
                ),
                name=f"dagster-daemon-{daemon_type}",
                daemon=True,  # Individual daemons should not outlive controller process
            )
            self._last_healthy_heartbeat_times[daemon_type] = time.time()
            self._daemon_threads[daemon_type].start()

        self._start_time = pendulum.now("UTC")

    def __enter__(self) -> Self:
        return self

    def _daemon_thread_healthy(self, daemon_type: str) -> bool:
        thread = self._daemon_threads[daemon_type]
        return thread.is_alive()

    def _daemon_heartbeat_health(self) -> Mapping[str, bool]:
        now = time.time()
        try:
            daemon_statuses_by_type = get_daemon_statuses(
                self._instance,
                daemon_types=self._daemons.keys(),
                heartbeat_interval_seconds=self._heartbeat_interval_seconds,
                heartbeat_tolerance_seconds=self._heartbeat_tolerance_seconds,
                ignore_errors=True,
            )
            daemon_health_by_type = {
                daemon_type: daemon_status.healthy
                for (daemon_type, daemon_status) in daemon_statuses_by_type.items()
            }

            for daemon_type, is_daemon_healthy in daemon_health_by_type.items():
                if is_daemon_healthy:
                    self._last_healthy_heartbeat_times[daemon_type] = now

            return daemon_health_by_type  # type: ignore  # (possible None)
        except Exception:
            self._logger.warning(
                "Error attempting to check daemon heartbeats",
                exc_info=sys.exc_info,  # type: ignore  # (should be func call)
            )

            return {
                daemon_type: (
                    self._last_healthy_heartbeat_times[daemon_type]
                    > now - self._heartbeat_tolerance_seconds
                )
                for daemon_type in self._daemons.keys()
            }

    def check_daemon_threads(self) -> None:
        failed_daemons = [
            daemon_type
            for daemon_type in self._daemon_threads
            if not self._daemon_thread_healthy(daemon_type)
        ]

        if failed_daemons:
            self._logger.error(
                "Stopping dagster-daemon process since the following threads are no longer"
                f" running: {failed_daemons}"
            )
            raise Exception("Stopped dagster-daemon process due to threads no longer running")

    def check_daemon_heartbeats(self) -> None:
        failed_daemons = [
            daemon_type
            for daemon_type, is_daemon_healthy in self._daemon_heartbeat_health().items()
            if not is_daemon_healthy
        ]

        if failed_daemons:
            self._logger.error(
                "Stopping dagster-daemon process since the following threads are no longer sending"
                f" heartbeats: {failed_daemons}"
            )
            raise Exception("Stopped dagster-daemon process due to thread heartbeat failure")

    def check_daemon_loop(self) -> None:
        start_time = time.time()
        last_heartbeat_check_time = start_time
        last_workspace_update_time = start_time
        while True:
            with raise_interrupts_as(KeyboardInterrupt):
                time.sleep(THREAD_CHECK_INTERVAL)
                self.check_daemon_threads()

                # periodically refresh the shared workspace context
                if (time.time() - last_workspace_update_time) > RELOAD_WORKSPACE_INTERVAL:
                    if self._grpc_server_registry:
                        self._grpc_server_registry.clear_all_grpc_endpoints()
                    self._workspace_process_context.refresh_workspace()
                    last_workspace_update_time = time.time()

                if self._instance.daemon_skip_heartbeats_without_errors:
                    # If we're skipping heartbeats without errors, we just check the threads.
                    # If there's no errors, the daemons won't be writing heartbeats.
                    continue

                now = time.time()
                # Give the daemon enough time to send an initial heartbeat before checking
                if (
                    (now - start_time) < 2 * self._heartbeat_interval_seconds
                    or now - last_heartbeat_check_time < HEARTBEAT_CHECK_INTERVAL
                ):
                    continue

                self.check_daemon_heartbeats()
                last_heartbeat_check_time = time.time()

    def __exit__(
        self,
        exception_type: Type[BaseException],
        exception_value: Exception,
        traceback: TracebackType,
    ) -> None:
        if isinstance(exception_value, KeyboardInterrupt):
            self._logger.info("Received interrupt, shutting down daemon threads...")
        elif exception_type:
            self._logger.warning(
                f"Shutting down daemon threads due to {exception_type.__name__}..."
            )
        else:
            self._logger.info("Shutting down daemon threads...")
        self._daemon_shutdown_event.set()
        for daemon_type, thread in self._daemon_threads.items():
            if thread.is_alive():
                thread.join(timeout=30)

                if thread.is_alive():
                    self._logger.error("Thread for %s did not shut down gracefully.", daemon_type)
        self._logger.info("Daemon threads shut down.")

    def _add_daemon(self, daemon: DagsterDaemon) -> None:
        self._daemons[daemon.daemon_type()] = daemon

    def get_daemon(self, daemon_type: str) -> DagsterDaemon:
        return self._daemons.get(daemon_type)  # type: ignore  # (possible none)

    @property
    def daemons(self) -> Sequence[DagsterDaemon]:
        return list(self._daemons.values())


def create_daemon_of_type(daemon_type: str, instance: DagsterInstance) -> DagsterDaemon:
    if daemon_type == SchedulerDaemon.daemon_type():
        return SchedulerDaemon()
    elif daemon_type == SensorDaemon.daemon_type():
        return SensorDaemon()
    elif daemon_type == QueuedRunCoordinatorDaemon.daemon_type():
        return QueuedRunCoordinatorDaemon(
            interval_seconds=instance.run_coordinator.dequeue_interval_seconds  # type: ignore  # (??)
        )
    elif daemon_type == BackfillDaemon.daemon_type():
        return BackfillDaemon(interval_seconds=DEFAULT_DAEMON_INTERVAL_SECONDS)
    elif daemon_type == MonitoringDaemon.daemon_type():
        return MonitoringDaemon(interval_seconds=instance.run_monitoring_poll_interval_seconds)
    elif daemon_type == EventLogConsumerDaemon.daemon_type():
        return EventLogConsumerDaemon()
    elif daemon_type == AssetDaemon.daemon_type():
        return AssetDaemon(
            interval_seconds=instance.auto_materialize_minimum_interval_seconds
            if instance.auto_materialize_minimum_interval_seconds is not None
            else DEFAULT_DAEMON_INTERVAL_SECONDS
        )
    else:
        raise Exception(f"Unexpected daemon type {daemon_type}")


def all_daemons_healthy(
    instance: DagsterInstance,
    curr_time_seconds: Optional[float] = None,
    heartbeat_interval_seconds: float = DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
    heartbeat_tolerance_seconds: float = DEFAULT_DAEMON_HEARTBEAT_TOLERANCE_SECONDS,
) -> bool:
    """True if all required daemons have had a recent heartbeat with no errors."""
    statuses_by_type = get_daemon_statuses(
        instance,
        daemon_types=instance.get_required_daemon_types(),
        heartbeat_interval_seconds=heartbeat_interval_seconds,
        heartbeat_tolerance_seconds=heartbeat_tolerance_seconds,
        curr_time_seconds=curr_time_seconds,
    )

    return all(status.healthy for status in statuses_by_type.values())


def all_daemons_live(
    instance: DagsterInstance,
    curr_time_seconds: Optional[float] = None,
    heartbeat_interval_seconds: float = DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
    heartbeat_tolerance_seconds: float = DEFAULT_DAEMON_HEARTBEAT_TOLERANCE_SECONDS,
) -> bool:
    """True if all required daemons have had a recent heartbeat, regardless of if it contained errors.
    """
    statuses_by_type = get_daemon_statuses(
        instance,
        daemon_types=instance.get_required_daemon_types(),
        heartbeat_interval_seconds=heartbeat_interval_seconds,
        heartbeat_tolerance_seconds=heartbeat_tolerance_seconds,
        curr_time_seconds=curr_time_seconds,
        ignore_errors=True,
    )

    return all(status.healthy for status in statuses_by_type.values())


def get_daemon_statuses(
    instance: DagsterInstance,
    daemon_types: Iterable[str],
    curr_time_seconds: Optional[float] = None,
    ignore_errors: bool = False,
    heartbeat_interval_seconds: float = DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
    heartbeat_tolerance_seconds: float = DEFAULT_DAEMON_HEARTBEAT_TOLERANCE_SECONDS,
) -> Mapping[str, DaemonStatus]:
    curr_time_seconds = check.opt_float_param(
        curr_time_seconds, "curr_time_seconds", default=pendulum.now("UTC").float_timestamp
    )

    daemon_statuses_by_type: Dict[str, DaemonStatus] = {}
    heartbeats = instance.get_daemon_heartbeats()

    for daemon_type in daemon_types:
        # check if daemon is not required
        if daemon_type not in instance.get_required_daemon_types():
            daemon_statuses_by_type[daemon_type] = DaemonStatus(
                daemon_type=daemon_type, required=False, healthy=None, last_heartbeat=None
            )
        else:
            # check if daemon has a heartbeat
            if daemon_type not in heartbeats:
                daemon_statuses_by_type[daemon_type] = DaemonStatus(
                    daemon_type=daemon_type, required=True, healthy=False, last_heartbeat=None
                )
            else:
                # check if daemon has sent a recent heartbeat
                latest_heartbeat = heartbeats[daemon_type]
                hearbeat_timestamp = latest_heartbeat.timestamp
                maximum_tolerated_time = (
                    hearbeat_timestamp + heartbeat_interval_seconds + heartbeat_tolerance_seconds
                )
                healthy = curr_time_seconds <= maximum_tolerated_time  # type: ignore  # (possible none)

                if not ignore_errors and latest_heartbeat.errors:
                    healthy = False

                daemon_statuses_by_type[daemon_type] = DaemonStatus(
                    daemon_type=daemon_type,
                    required=True,
                    healthy=healthy,
                    last_heartbeat=heartbeats[daemon_type],
                )

    return daemon_statuses_by_type


def debug_daemon_heartbeats(instance: DagsterInstance) -> None:
    daemon = SensorDaemon()
    timestamp = pendulum.now("UTC").float_timestamp
    instance.add_daemon_heartbeat(DaemonHeartbeat(timestamp, daemon.daemon_type(), None, None))
    returned_timestamp = instance.get_daemon_heartbeats()[daemon.daemon_type()].timestamp
    print(f"Written timestamp: {timestamp}\nRead timestamp: {returned_timestamp}")  # noqa: T201



"""add instigators table.

Revision ID: 5b467f7af3f6
Revises: 54666da3db5c
Create Date: 2022-03-18 16:17:20.338259

"""
from dagster._core.storage.migration.utils import create_instigators_table

# revision identifiers, used by Alembic.
revision = "5b467f7af3f6"
down_revision = "54666da3db5c"
branch_labels = None
depends_on = None


def upgrade():
    create_instigators_table()


def downgrade():
    pass



from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import IO, Callable, Generator, Iterator, NamedTuple, Optional, Sequence

from typing_extensions import Final, Self

import dagster._check as check
from dagster._core.storage.compute_log_manager import ComputeIOType

MAX_BYTES_CHUNK_READ: Final = 4194304  # 4 MB


class CapturedLogContext(
    NamedTuple(
        "_CapturedLogContext",
        [
            ("log_key", Sequence[str]),
            ("external_url", Optional[str]),
            ("external_stdout_url", Optional[str]),
            ("external_stderr_url", Optional[str]),
        ],
    )
):
    """Object representing the context in which logs are captured.  Can be used by external logging
    sidecar implementations to point dagit to an external url to view compute logs instead of a
    Dagster-managed location.
    """

    def __new__(
        cls,
        log_key: Sequence[str],
        external_stdout_url: Optional[str] = None,
        external_stderr_url: Optional[str] = None,
        external_url: Optional[str] = None,
    ):
        if external_url and (external_stdout_url or external_stderr_url):
            check.failed(
                "Cannot specify both `external_url` and one of"
                " `external_stdout_url`/`external_stderr_url`"
            )

        return super(CapturedLogContext, cls).__new__(
            cls,
            log_key,
            external_stdout_url=external_stdout_url,
            external_stderr_url=external_stderr_url,
            external_url=external_url,
        )


class CapturedLogData(
    NamedTuple(
        "_CapturedLogData",
        [
            ("log_key", Sequence[str]),
            ("stdout", Optional[bytes]),
            ("stderr", Optional[bytes]),
            ("cursor", Optional[str]),
        ],
    )
):
    """Object representing captured log data, either a partial chunk of the log data or the full
    capture.  Contains the raw bytes and optionally the cursor offset for the partial chunk.
    """

    def __new__(
        cls,
        log_key: Sequence[str],
        stdout: Optional[bytes] = None,
        stderr: Optional[bytes] = None,
        cursor: Optional[str] = None,
    ):
        return super(CapturedLogData, cls).__new__(cls, log_key, stdout, stderr, cursor)


class CapturedLogMetadata(
    NamedTuple(
        "_CapturedLogMetadata",
        [
            ("stdout_location", Optional[str]),
            ("stderr_location", Optional[str]),
            ("stdout_download_url", Optional[str]),
            ("stderr_download_url", Optional[str]),
        ],
    )
):
    """Object representing metadata info for the captured log data, containing a display string for
    the location of the log data and a URL for direct download of the captured log data.
    """

    def __new__(
        cls,
        stdout_location: Optional[str] = None,
        stderr_location: Optional[str] = None,
        stdout_download_url: Optional[str] = None,
        stderr_download_url: Optional[str] = None,
    ):
        return super(CapturedLogMetadata, cls).__new__(
            cls,
            stdout_location=stdout_location,
            stderr_location=stderr_location,
            stdout_download_url=stdout_download_url,
            stderr_download_url=stderr_download_url,
        )


class CapturedLogSubscription:
    def __init__(self, manager: CapturedLogManager, log_key: Sequence[str], cursor: Optional[str]):
        self._manager = manager
        self._log_key = log_key
        self._cursor = cursor
        self._observer: Optional[Callable[[CapturedLogData], None]] = None
        self.is_complete = False

    def __call__(self, observer: Optional[Callable[[CapturedLogData], None]]) -> Self:
        self._observer = observer
        self.fetch()
        if self._manager.is_capture_complete(self._log_key):
            self.complete()
        return self

    @property
    def log_key(self) -> Sequence[str]:
        return self._log_key

    def dispose(self) -> None:
        self._observer = None
        self._manager.unsubscribe(self)

    def fetch(self) -> None:
        if not self._observer:
            return

        should_fetch = True
        while should_fetch:
            log_data = self._manager.get_log_data(
                self._log_key,
                self._cursor,
                max_bytes=MAX_BYTES_CHUNK_READ,
            )
            if not self._cursor or log_data.cursor != self._cursor:
                self._observer(log_data)
                self._cursor = log_data.cursor
            should_fetch = _has_max_data(log_data.stdout) or _has_max_data(log_data.stderr)

    def complete(self) -> None:
        self.is_complete = True


def _has_max_data(chunk: Optional[bytes]) -> bool:
    # function is used as predicate but does not actually return a boolean
    return chunk and len(chunk) >= MAX_BYTES_CHUNK_READ  # type: ignore


class CapturedLogManager(ABC):
    """Abstract base class for capturing the unstructured logs (stdout/stderr) in the current
    process, stored / retrieved with a provided log_key.
    """

    @abstractmethod
    @contextmanager
    def capture_logs(self, log_key: Sequence[str]) -> Generator[CapturedLogContext, None, None]:
        """Context manager for capturing the stdout/stderr within the current process, and persisting
        it under the given log key.

        Args:
            log_key (List[String]): The log key identifying the captured logs
        """

    @abstractmethod
    @contextmanager
    def open_log_stream(
        self, log_key: Sequence[str], io_type: ComputeIOType
    ) -> Iterator[Optional[IO[bytes]]]:
        """Context manager for providing an IO stream that enables the caller to write to a log stream
        managed by the captured log manager, to be read later using the given log key.

        Args:
            log_key (List[String]): The log key identifying the captured logs
        """

    @abstractmethod
    def is_capture_complete(self, log_key: Sequence[str]) -> bool:
        """Flag indicating when the log capture for a given log key has completed.

        Args:
            log_key (List[String]): The log key identifying the captured logs

        Returns:
            Boolean
        """

    @abstractmethod
    def get_log_data(
        self,
        log_key: Sequence[str],
        cursor: Optional[str] = None,
        max_bytes: Optional[int] = None,
    ) -> CapturedLogData:
        """Returns a chunk of the captured stdout logs for a given log key.

        Args:
            log_key (List[String]): The log key identifying the captured logs
            cursor (Optional[str]): A cursor representing the position of the log chunk to fetch
            max_bytes (Optional[int]): A limit on the size of the log chunk to fetch

        Returns:
            CapturedLogData
        """

    @abstractmethod
    def get_log_metadata(self, log_key: Sequence[str]) -> CapturedLogMetadata:
        """Returns the metadata of the captured logs for a given log key, including
        displayable information on where the logs are persisted.

        Args:
            log_key (List[String]): The log key identifying the captured logs

        Returns:
            CapturedLogMetadata
        """

    @abstractmethod
    def delete_logs(
        self, log_key: Optional[Sequence[str]] = None, prefix: Optional[Sequence[str]] = None
    ) -> None:
        """Deletes the captured logs for a given log key.

        Args:
            log_key(Optional[List[String]]): The log key of the logs to delete
            prefix(Optional[List[String]]): The prefix of the log keys to delete
        """

    @abstractmethod
    def subscribe(
        self, log_key: Sequence[str], cursor: Optional[str] = None
    ) -> CapturedLogSubscription:
        """Registers an observable object for log data.

        Args:
            log_key (List[String]): The log key identifying the captured logs
            cursor (Optional[String]): The string cursor marking the position within the log stream
        Returns:
            ComputeLogSubscription
        """

    @abstractmethod
    def unsubscribe(self, subscription: CapturedLogSubscription) -> None:
        """Deregisters an observable object from receiving log updates.

        Args:
            subscription (CapturedLogSubscription): subscription object which manages when to send
                back data to the subscriber
        """

    def build_log_key_for_run(self, run_id: str, step_key: str) -> Sequence[str]:
        """Legacy adapter to translate run_id/key to captured log manager-based log_key."""
        return [run_id, "compute_logs", step_key]



from typing import Dict, Tuple

from dagster._core.execution.context.input import InputContext
from dagster._core.execution.context.output import OutputContext
from dagster._core.storage.io_manager import IOManager, dagster_maintained_io_manager, io_manager


class InMemoryIOManager(IOManager):
    def __init__(self):
        self.values: Dict[Tuple[object, ...], object] = {}

    def handle_output(self, context: OutputContext, obj: object):
        keys = tuple(context.get_identifier())
        self.values[keys] = obj

    def load_input(self, context: InputContext) -> object:
        keys = tuple(context.get_identifier())
        return self.values[keys]


@dagster_maintained_io_manager
@io_manager(description="Built-in IO manager that stores and retrieves values in memory.")
def mem_io_manager(_) -> InMemoryIOManager:
    """Built-in IO manager that stores and retrieves values in memory."""
    return InMemoryIOManager()



import importlib
import os
import warnings
from datetime import datetime
from functools import update_wrapper
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

from typing_extensions import Self

import dagster._check as check
from dagster._annotations import public
from dagster._config import Field, Shape, StringSource
from dagster._config.config_type import ConfigType
from dagster._config.validate import validate_config
from dagster._core.definitions.dependency import (
    Node,
    NodeHandle,
    NodeInputHandle,
    NodeInvocation,
)
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.node_definition import NodeDefinition
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.op_selection import OpSelection, get_graph_subset
from dagster._core.definitions.partition import DynamicPartitionsDefinition
from dagster._core.definitions.policy import RetryPolicy
from dagster._core.definitions.resource_requirement import (
    ResourceRequirement,
    ensure_requirements_satisfied,
)
from dagster._core.definitions.utils import check_valid_name
from dagster._core.errors import (
    DagsterInvalidConfigError,
    DagsterInvalidDefinitionError,
    DagsterInvalidInvocationError,
    DagsterInvalidSubsetError,
    DagsterInvariantViolationError,
)
from dagster._core.selector.subset_selector import (
    AssetSelectionData,
    OpSelectionData,
)
from dagster._core.storage.io_manager import (
    IOManagerDefinition,
    dagster_maintained_io_manager,
    io_manager,
)
from dagster._core.storage.tags import MEMOIZED_RUN_TAG
from dagster._core.types.dagster_type import DagsterType
from dagster._core.utils import str_format_set
from dagster._utils import IHasInternalInit
from dagster._utils.backcompat import deprecation_warning, experimental_class_warning
from dagster._utils.merger import merge_dicts

from .asset_layer import AssetLayer, build_asset_selection_job
from .config import ConfigMapping
from .dependency import (
    DependencyMapping,
    DependencyStructure,
    OpNode,
)
from .executor_definition import ExecutorDefinition, multi_or_in_process_executor
from .graph_definition import GraphDefinition, SubselectedGraphDefinition
from .hook_definition import HookDefinition
from .logger_definition import LoggerDefinition
from .metadata import MetadataValue, RawMetadataValue, normalize_metadata
from .partition import PartitionedConfig, PartitionsDefinition
from .resource_definition import ResourceDefinition
from .run_request import RunRequest
from .utils import DEFAULT_IO_MANAGER_KEY, validate_tags
from .version_strategy import VersionStrategy

if TYPE_CHECKING:
    from dagster._config.snap import ConfigSchemaSnapshot
    from dagster._core.definitions.run_config import RunConfig
    from dagster._core.execution.execute_in_process_result import ExecuteInProcessResult
    from dagster._core.execution.resources_init import InitResourceContext
    from dagster._core.host_representation.job_index import JobIndex
    from dagster._core.instance import DagsterInstance
    from dagster._core.snap import JobSnapshot

    from .run_config_schema import RunConfigSchema

DEFAULT_EXECUTOR_DEF = multi_or_in_process_executor


class JobDefinition(IHasInternalInit):
    """Defines a Dagster job."""

    _name: str
    _graph_def: GraphDefinition
    _description: Optional[str]
    _tags: Mapping[str, str]
    _metadata: Mapping[str, MetadataValue]
    _current_level_node_defs: Sequence[NodeDefinition]
    _hook_defs: AbstractSet[HookDefinition]
    _op_retry_policy: Optional[RetryPolicy]
    _asset_layer: AssetLayer
    _resource_requirements: Mapping[str, AbstractSet[str]]
    _all_node_defs: Mapping[str, NodeDefinition]
    _cached_run_config_schemas: Dict[str, "RunConfigSchema"]
    _version_strategy: VersionStrategy
    _subset_selection_data: Optional[Union[OpSelectionData, AssetSelectionData]]
    input_values: Mapping[str, object]

    def __init__(
        self,
        *,
        graph_def: GraphDefinition,
        resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
        executor_def: Optional[ExecutorDefinition] = None,
        logger_defs: Optional[Mapping[str, LoggerDefinition]] = None,
        name: Optional[str] = None,
        config: Optional[
            Union[ConfigMapping, Mapping[str, object], PartitionedConfig, "RunConfig"]
        ] = None,
        description: Optional[str] = None,
        partitions_def: Optional[PartitionsDefinition] = None,
        tags: Optional[Mapping[str, Any]] = None,
        metadata: Optional[Mapping[str, RawMetadataValue]] = None,
        hook_defs: Optional[AbstractSet[HookDefinition]] = None,
        op_retry_policy: Optional[RetryPolicy] = None,
        version_strategy: Optional[VersionStrategy] = None,
        _subset_selection_data: Optional[Union[OpSelectionData, AssetSelectionData]] = None,
        asset_layer: Optional[AssetLayer] = None,
        input_values: Optional[Mapping[str, object]] = None,
        _was_explicitly_provided_resources: Optional[bool] = None,
    ):
        from dagster._core.definitions.run_config import RunConfig, convert_config_input

        self._graph_def = graph_def
        self._current_level_node_defs = self._graph_def.node_defs
        # Recursively explore all nodes in the this job
        self._all_node_defs = _build_all_node_defs(self._current_level_node_defs)
        self._asset_layer = check.opt_inst_param(
            asset_layer, "asset_layer", AssetLayer
        ) or _infer_asset_layer_from_source_asset_deps(graph_def)

        # validates
        self._graph_def.get_inputs_must_be_resolved_top_level(self._asset_layer)

        self._name = check_valid_name(check.str_param(name, "name")) if name else graph_def.name
        self._executor_def = check.opt_inst_param(executor_def, "executor_def", ExecutorDefinition)
        self._loggers = check.opt_nullable_mapping_param(
            logger_defs,
            "logger_defs",
            key_type=str,
            value_type=LoggerDefinition,
        )

        config = check.opt_inst_param(
            config, "config", (Mapping, ConfigMapping, PartitionedConfig, RunConfig)
        )
        config = convert_config_input(config)

        partitions_def = check.opt_inst_param(
            partitions_def, "partitions_def", PartitionsDefinition
        )
        # tags and description can exist on graph as well, but since
        # same graph may be in multiple jobs, keep separate layer
        self._description = check.opt_str_param(description, "description")
        self._tags = validate_tags(tags)
        self._metadata = normalize_metadata(
            check.opt_mapping_param(metadata, "metadata", key_type=str)
        )
        self._hook_defs = check.opt_set_param(hook_defs, "hook_defs")
        self._op_retry_policy = check.opt_inst_param(
            op_retry_policy, "op_retry_policy", RetryPolicy
        )
        self.version_strategy = check.opt_inst_param(
            version_strategy, "version_strategy", VersionStrategy
        )
        if self.version_strategy is not None:
            experimental_class_warning("VersionStrategy")

        _subset_selection_data = check.opt_inst_param(
            _subset_selection_data, "_subset_selection_data", (OpSelectionData, AssetSelectionData)
        )
        input_values = check.opt_mapping_param(input_values, "input_values", key_type=str)

        resource_defs = check.opt_mapping_param(
            resource_defs, "resource_defs", key_type=str, value_type=ResourceDefinition
        )
        for key in resource_defs.keys():
            if not key.isidentifier():
                check.failed(f"Resource key '{key}' must be a valid Python identifier.")
        was_provided_resources = (
            bool(resource_defs)
            if _was_explicitly_provided_resources is None
            else _was_explicitly_provided_resources
        )
        self._resource_defs = {
            DEFAULT_IO_MANAGER_KEY: default_job_io_manager,
            **resource_defs,
        }
        self._required_resource_keys = self._get_required_resource_keys(was_provided_resources)

        self._config_mapping = None
        self._partitioned_config = None
        self._run_config = None
        self._run_config_schema = None
        self._original_config_argument = config

        if partitions_def:
            self._partitioned_config = PartitionedConfig.from_flexible_config(
                config, partitions_def
            )
        else:
            if isinstance(config, ConfigMapping):
                self._config_mapping = config
            elif isinstance(config, PartitionedConfig):
                self._partitioned_config = config
            elif isinstance(config, dict):
                self._run_config = config
                # Using config mapping here is a trick to make it so that the preset will be used even
                # when no config is supplied for the job.
                self._config_mapping = _config_mapping_with_default_value(
                    get_run_config_schema_for_job(
                        graph_def,
                        self.resource_defs,
                        self.executor_def,
                        self.loggers,
                        asset_layer,
                        was_explicitly_provided_resources=was_provided_resources,
                    ),
                    config,
                    self.name,
                )
            elif config is not None:
                check.failed(
                    "config param must be a ConfigMapping, a PartitionedConfig, or a dictionary,"
                    f" but is an object of type {type(config)}"
                )

        self._subset_selection_data = _subset_selection_data
        self.input_values = input_values
        for input_name in sorted(list(self.input_values.keys())):
            if not graph_def.has_input(input_name):
                raise DagsterInvalidDefinitionError(
                    f"Error when constructing JobDefinition '{self.name}': Input value provided for"
                    f" key '{input_name}', but job has no top-level input with that name."
                )

    def dagster_internal_init(
        *,
        graph_def: GraphDefinition,
        resource_defs: Optional[Mapping[str, ResourceDefinition]],
        executor_def: Optional[ExecutorDefinition],
        logger_defs: Optional[Mapping[str, LoggerDefinition]],
        name: Optional[str],
        config: Optional[
            Union[ConfigMapping, Mapping[str, object], PartitionedConfig, "RunConfig"]
        ],
        description: Optional[str],
        partitions_def: Optional[PartitionsDefinition],
        tags: Optional[Mapping[str, Any]],
        metadata: Optional[Mapping[str, RawMetadataValue]],
        hook_defs: Optional[AbstractSet[HookDefinition]],
        op_retry_policy: Optional[RetryPolicy],
        version_strategy: Optional[VersionStrategy],
        _subset_selection_data: Optional[Union[OpSelectionData, AssetSelectionData]],
        asset_layer: Optional[AssetLayer],
        input_values: Optional[Mapping[str, object]],
        _was_explicitly_provided_resources: Optional[bool],
    ) -> "JobDefinition":
        return JobDefinition(
            graph_def=graph_def,
            resource_defs=resource_defs,
            executor_def=executor_def,
            logger_defs=logger_defs,
            name=name,
            config=config,
            description=description,
            partitions_def=partitions_def,
            tags=tags,
            metadata=metadata,
            hook_defs=hook_defs,
            op_retry_policy=op_retry_policy,
            version_strategy=version_strategy,
            _subset_selection_data=_subset_selection_data,
            asset_layer=asset_layer,
            input_values=input_values,
            _was_explicitly_provided_resources=_was_explicitly_provided_resources,
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def tags(self) -> Mapping[str, str]:
        return merge_dicts(self._graph_def.tags, self._tags)

    @property
    def metadata(self) -> Mapping[str, MetadataValue]:
        return self._metadata

    @property
    def description(self) -> Optional[str]:
        return self._description

    @property
    def graph(self) -> GraphDefinition:
        return self._graph_def

    @property
    def dependency_structure(self) -> DependencyStructure:
        return self._graph_def.dependency_structure

    @property
    def dependencies(self) -> DependencyMapping[NodeInvocation]:
        return self._graph_def.dependencies

    @public
    @property
    def executor_def(self) -> ExecutorDefinition:
        return self._executor_def or DEFAULT_EXECUTOR_DEF

    @public
    @property
    def has_specified_executor(self) -> bool:
        return self._executor_def is not None

    @public
    @property
    def resource_defs(self) -> Mapping[str, ResourceDefinition]:
        return self._resource_defs

    @public
    @property
    def partitioned_config(self) -> Optional[PartitionedConfig]:
        return self._partitioned_config

    @public
    @property
    def config_mapping(self) -> Optional[ConfigMapping]:
        return self._config_mapping

    @public
    @property
    def loggers(self) -> Mapping[str, LoggerDefinition]:
        from dagster._loggers import default_loggers

        return self._loggers or default_loggers()

    @public
    @property
    def has_specified_loggers(self) -> bool:
        return self._loggers is not None

    @property
    def required_resource_keys(self) -> AbstractSet[str]:
        return self._required_resource_keys

    @property
    def run_config(self) -> Optional[Mapping[str, Any]]:
        return self._run_config

    @property
    def run_config_schema(self) -> "RunConfigSchema":
        if self._run_config_schema is None:
            self._run_config_schema = _create_run_config_schema(self, self.required_resource_keys)
        return self._run_config_schema

    @public
    @property
    def partitions_def(self) -> Optional[PartitionsDefinition]:
        return None if not self.partitioned_config else self.partitioned_config.partitions_def

    @property
    def hook_defs(self) -> AbstractSet[HookDefinition]:
        return self._hook_defs

    @property
    def asset_layer(self) -> AssetLayer:
        return self._asset_layer

    @property
    def all_node_defs(self) -> Sequence[NodeDefinition]:
        return list(self._all_node_defs.values())

    @property
    def top_level_node_defs(self) -> Sequence[NodeDefinition]:
        return self._current_level_node_defs

    def node_def_named(self, name: str) -> NodeDefinition:
        check.str_param(name, "name")

        check.invariant(name in self._all_node_defs, f"{name} not found")
        return self._all_node_defs[name]

    def has_node(self, name: str) -> bool:
        check.str_param(name, "name")
        return name in self._all_node_defs

    def get_node(self, handle: NodeHandle) -> Node:
        return self._graph_def.get_node(handle)

    def get_op(self, handle: NodeHandle) -> OpNode:
        node = self.get_node(handle)
        assert isinstance(
            node, OpNode
        ), f"Tried to retrieve node {handle} as op, but it represents a nested graph."
        return node

    def has_node_named(self, name: str) -> bool:
        return self._graph_def.has_node_named(name)

    def get_node_named(self, name: str) -> Node:
        return self._graph_def.node_named(name)

    @property
    def nodes(self) -> Sequence[Node]:
        return self._graph_def.nodes

    @property
    def nodes_in_topological_order(self) -> Sequence[Node]:
        return self._graph_def.nodes_in_topological_order

    def all_dagster_types(self) -> Iterable[DagsterType]:
        return self._graph_def.all_dagster_types()

    def has_dagster_type(self, name: str) -> bool:
        return self._graph_def.has_dagster_type(name)

    def dagster_type_named(self, name: str) -> DagsterType:
        return self._graph_def.dagster_type_named(name)

    def describe_target(self) -> str:
        return f"job '{self.name}'"

    def is_using_memoization(self, run_tags: Mapping[str, str]) -> bool:
        tags = merge_dicts(self.tags, run_tags)
        # If someone provides a false value for memoized run tag, then they are intentionally
        # switching off memoization.
        if tags.get(MEMOIZED_RUN_TAG) == "false":
            return False
        return (
            MEMOIZED_RUN_TAG in tags and tags.get(MEMOIZED_RUN_TAG) == "true"
        ) or self.version_strategy is not None

    def get_required_resource_defs(self) -> Mapping[str, ResourceDefinition]:
        return {
            resource_key: resource
            for resource_key, resource in self.resource_defs.items()
            if resource_key in self.required_resource_keys
        }

    def _get_required_resource_keys(self, validate_requirements: bool = False) -> AbstractSet[str]:
        from ..execution.resources_init import get_transitive_required_resource_keys

        requirements = self._get_resource_requirements()
        if validate_requirements:
            ensure_requirements_satisfied(self.resource_defs, requirements)
        required_keys = {req.key for req in requirements}
        if validate_requirements:
            return required_keys.union(
                get_transitive_required_resource_keys(required_keys, self.resource_defs)
            )
        else:
            return required_keys

    def _get_resource_requirements(self) -> Sequence[ResourceRequirement]:
        return [
            *self._graph_def.get_resource_requirements(self.asset_layer),
            *[
                req
                for hook_def in self._hook_defs
                for req in hook_def.get_resource_requirements(outer_context=f"job '{self._name}'")
            ],
        ]

    def validate_resource_requirements_satisfied(self) -> None:
        resource_requirements = self._get_resource_requirements()
        ensure_requirements_satisfied(self.resource_defs, resource_requirements)

    def is_missing_required_resources(self) -> bool:
        requirements = self._get_resource_requirements()
        for requirement in requirements:
            if not requirement.resources_contain_key(self.resource_defs):
                return True
        return False

    def get_all_hooks_for_handle(self, handle: NodeHandle) -> AbstractSet[HookDefinition]:
        """Gather all the hooks for the given node from all places possibly attached with a hook.

        A hook can be attached to any of the following objects
        * Node (node invocation)
        * JobDefinition

        Args:
            handle (NodeHandle): The node's handle

        Returns:
            FrozenSet[HookDefinition]
        """
        check.inst_param(handle, "handle", NodeHandle)
        hook_defs: Set[HookDefinition] = set()

        current = handle
        lineage = []
        while current:
            lineage.append(current.name)
            current = current.parent

        # hooks on top-level node
        name = lineage.pop()
        node = self._graph_def.node_named(name)
        hook_defs = hook_defs.union(node.hook_defs)

        # hooks on non-top-level nodes
        while lineage:
            name = lineage.pop()
            # While lineage is non-empty, definition is guaranteed to be a graph
            definition = cast(GraphDefinition, node.definition)
            node = definition.node_named(name)
            hook_defs = hook_defs.union(node.hook_defs)

        # hooks applied to a job definition will run on every node
        hook_defs = hook_defs.union(self.hook_defs)

        return frozenset(hook_defs)

    def get_retry_policy_for_handle(self, handle: NodeHandle) -> Optional[RetryPolicy]:
        node = self.get_node(handle)
        definition = node.definition

        if node.retry_policy:
            return node.retry_policy
        elif isinstance(definition, OpDefinition) and definition.retry_policy:
            return definition.retry_policy

        # could be expanded to look in graph containers
        else:
            return self._op_retry_policy

    # make Callable for decorator reference updates
    def __call__(self, *args, **kwargs):
        raise DagsterInvariantViolationError(
            f"Attempted to call job '{self.name}' directly. Jobs should be invoked by "
            "using an execution API function (e.g. `job.execute_in_process`)."
        )

    @public
    def execute_in_process(
        self,
        run_config: Optional[Union[Mapping[str, Any], "RunConfig"]] = None,
        instance: Optional["DagsterInstance"] = None,
        partition_key: Optional[str] = None,
        raise_on_error: bool = True,
        op_selection: Optional[Sequence[str]] = None,
        asset_selection: Optional[Sequence[AssetKey]] = None,
        run_id: Optional[str] = None,
        input_values: Optional[Mapping[str, object]] = None,
        tags: Optional[Mapping[str, str]] = None,
        resources: Optional[Mapping[str, object]] = None,
    ) -> "ExecuteInProcessResult":
        """Execute the Job in-process, gathering results in-memory.

        The `executor_def` on the Job will be ignored, and replaced with the in-process executor.
        If using the default `io_manager`, it will switch from filesystem to in-memory.


        Args:
            run_config (Optional[Mapping[str, Any]]:
                The configuration for the run
            instance (Optional[DagsterInstance]):
                The instance to execute against, an ephemeral one will be used if none provided.
            partition_key: (Optional[str])
                The string partition key that specifies the run config to execute. Can only be used
                to select run config for jobs with partitioned config.
            raise_on_error (Optional[bool]): Whether or not to raise exceptions when they occur.
                Defaults to ``True``.
            op_selection (Optional[Sequence[str]]): A list of op selection queries (including single op
                names) to execute. For example:
                * ``['some_op']``: selects ``some_op`` itself.
                * ``['*some_op']``: select ``some_op`` and all its ancestors (upstream dependencies).
                * ``['*some_op+++']``: select ``some_op``, all its ancestors, and its descendants
                (downstream dependencies) within 3 levels down.
                * ``['*some_op', 'other_op_a', 'other_op_b+']``: select ``some_op`` and all its
                ancestors, ``other_op_a`` itself, and ``other_op_b`` and its direct child ops.
            input_values (Optional[Mapping[str, Any]]):
                A dictionary that maps python objects to the top-level inputs of the job. Input values provided here will override input values that have been provided to the job directly.
            resources (Optional[Mapping[str, Any]]):
                The resources needed if any are required. Can provide resource instances directly,
                or resource definitions.

        Returns:
            :py:class:`~dagster.ExecuteInProcessResult`

        """
        from dagster._core.definitions.executor_definition import execute_in_process_executor
        from dagster._core.definitions.run_config import convert_config_input
        from dagster._core.execution.build_resources import wrap_resources_for_execution
        from dagster._core.execution.execute_in_process import core_execute_in_process

        run_config = check.opt_mapping_param(convert_config_input(run_config), "run_config")
        op_selection = check.opt_sequence_param(op_selection, "op_selection", str)
        asset_selection = check.opt_sequence_param(asset_selection, "asset_selection", AssetKey)
        resources = check.opt_mapping_param(resources, "resources", key_type=str)

        resource_defs = wrap_resources_for_execution(resources)

        check.invariant(
            not (op_selection and asset_selection),
            (
                "op_selection and asset_selection cannot both be provided as args to"
                " execute_in_process"
            ),
        )

        partition_key = check.opt_str_param(partition_key, "partition_key")
        input_values = check.opt_mapping_param(input_values, "input_values")

        # Combine provided input values at execute_in_process with input values
        # provided to the definition. Input values provided at
        # execute_in_process will override those provided on the definition.
        input_values = merge_dicts(self.input_values, input_values)

        bound_resource_defs = dict(self.resource_defs)
        ephemeral_job = JobDefinition.dagster_internal_init(
            name=self._name,
            graph_def=self._graph_def,
            resource_defs={**_swap_default_io_man(bound_resource_defs, self), **resource_defs},
            executor_def=execute_in_process_executor,
            logger_defs=self._loggers,
            hook_defs=self.hook_defs,
            config=self.config_mapping or self.partitioned_config or self.run_config,
            tags=self.tags,
            op_retry_policy=self._op_retry_policy,
            version_strategy=self.version_strategy,
            asset_layer=self.asset_layer,
            input_values=input_values,
            description=self.description,
            partitions_def=self.partitions_def,
            metadata=self.metadata,
            _subset_selection_data=None,  # this is added below
            _was_explicitly_provided_resources=True,
        )

        ephemeral_job = ephemeral_job.get_subset(
            op_selection=op_selection,
            asset_selection=frozenset(asset_selection) if asset_selection else None,
        )

        merged_tags = merge_dicts(self.tags, tags or {})
        if partition_key:
            if not (self.partitions_def and self.partitioned_config):
                check.failed("Attempted to execute a partitioned run for a non-partitioned job")
            self.partitions_def.validate_partition_key(
                partition_key, dynamic_partitions_store=instance
            )

            run_config = (
                run_config
                if run_config
                else self.partitioned_config.get_run_config_for_partition_key(partition_key)
            )
            merged_tags.update(
                self.partitioned_config.get_tags_for_partition_key(
                    partition_key, job_name=self.name
                )
            )

        return core_execute_in_process(
            ephemeral_job=ephemeral_job,
            run_config=run_config,
            instance=instance,
            output_capturing_enabled=True,
            raise_on_error=raise_on_error,
            run_tags=merged_tags,
            run_id=run_id,
            asset_selection=frozenset(asset_selection),
        )

    @property
    def op_selection_data(self) -> Optional[OpSelectionData]:
        return (
            self._subset_selection_data
            if isinstance(self._subset_selection_data, OpSelectionData)
            else None
        )

    @property
    def asset_selection_data(self) -> Optional[AssetSelectionData]:
        return (
            self._subset_selection_data
            if isinstance(self._subset_selection_data, AssetSelectionData)
            else None
        )

    @property
    def is_subset(self) -> bool:
        return bool(self._subset_selection_data)

    def get_subset(
        self,
        *,
        op_selection: Optional[Iterable[str]] = None,
        asset_selection: Optional[AbstractSet[AssetKey]] = None,
    ) -> Self:
        check.invariant(
            not (op_selection and asset_selection),
            (
                "op_selection and asset_selection cannot both be provided as args to"
                " execute_in_process"
            ),
        )
        if op_selection:
            return self._get_job_def_for_op_selection(op_selection)
        if asset_selection:
            return self._get_job_def_for_asset_selection(asset_selection)
        else:
            return self

    def _get_job_def_for_asset_selection(
        self,
        asset_selection: Optional[AbstractSet[AssetKey]] = None,
    ) -> Self:
        asset_selection = check.opt_set_param(asset_selection, "asset_selection", AssetKey)

        nonexistent_assets = [
            asset
            for asset in asset_selection
            if asset not in self.asset_layer.asset_keys
            and asset not in self.asset_layer.source_assets_by_key
        ]
        nonexistent_asset_strings = [
            asset_str
            for asset_str in (asset.to_string() for asset in nonexistent_assets)
            if asset_str
        ]
        if nonexistent_assets:
            raise DagsterInvalidSubsetError(
                "Assets provided in asset_selection argument "
                f"{', '.join(nonexistent_asset_strings)} do not exist in parent asset group or job."
            )
        asset_selection_data = AssetSelectionData(
            asset_selection=asset_selection,
            parent_job_def=self,
        )

        check.invariant(
            self.asset_layer.assets_defs_by_key is not None,
            "Asset layer must have _asset_defs argument defined",
        )

        new_job = build_asset_selection_job(
            name=self.name,
            assets=set(self.asset_layer.assets_defs_by_key.values()),
            source_assets=self.asset_layer.source_assets_by_key.values(),
            executor_def=self.executor_def,
            resource_defs=self.resource_defs,
            description=self.description,
            tags=self.tags,
            asset_selection=asset_selection,
            asset_selection_data=asset_selection_data,
            config=self.config_mapping or self.partitioned_config,
        )
        return new_job

    def _get_job_def_for_op_selection(self, op_selection: Iterable[str]) -> Self:
        try:
            sub_graph = get_graph_subset(self.graph, op_selection)

            # if explicit config was passed the config_mapping that resolves the defaults implicitly is
            # very unlikely to work. The job will still present the default config in dagit.
            config = (
                None
                if self.run_config is not None
                else self.config_mapping or self.partitioned_config
            )

            return self._copy(
                config=config,
                graph_def=sub_graph,
                _subset_selection_data=OpSelectionData(
                    op_selection=list(op_selection),
                    resolved_op_selection=OpSelection(op_selection).resolve(self.graph),
                    parent_job_def=self,  # used by job snapshot lineage
                ),
                # TODO: subset this structure.
                # https://github.com/dagster-io/dagster/issues/7541
                asset_layer=self.asset_layer,
            )
        except DagsterInvalidDefinitionError as exc:
            # This handles the case when you construct a subset such that an unsatisfied
            # input cannot be loaded from config. Instead of throwing a DagsterInvalidDefinitionError,
            # we re-raise a DagsterInvalidSubsetError.
            node_paths = OpSelection(op_selection).resolve(self.graph)
            raise DagsterInvalidSubsetError(
                f"The attempted subset {str_format_set(node_paths)} for graph "
                f"{self.graph.name} results in an invalid graph."
            ) from exc

    @public
    def run_request_for_partition(
        self,
        partition_key: str,
        run_key: Optional[str] = None,
        tags: Optional[Mapping[str, str]] = None,
        asset_selection: Optional[Sequence[AssetKey]] = None,
        run_config: Optional[Mapping[str, Any]] = None,
        current_time: Optional[datetime] = None,
    ) -> RunRequest:
        """Creates a RunRequest object for a run that processes the given partition.

        Args:
            partition_key: The key of the partition to request a run for.
            run_key (Optional[str]): A string key to identify this launched run. For sensors, ensures that
                only one run is created per run key across all sensor evaluations.  For schedules,
                ensures that one run is created per tick, across failure recoveries. Passing in a `None`
                value means that a run will always be launched per evaluation.
            tags (Optional[Dict[str, str]]): A dictionary of tags (string key-value pairs) to attach
                to the launched run.
            run_config (Optional[Mapping[str, Any]]: Configuration for the run. If the job has
                a :py:class:`PartitionedConfig`, this value will override replace the config
                provided by it.
            current_time (Optional[datetime): Used to determine which time-partitions exist.
                Defaults to now.

        Returns:
            RunRequest: an object that requests a run to process the given partition.
        """
        deprecation_warning(
            "JobDefinition.run_request_for_partition",
            "2.0.0",
            additional_warn_txt="Directly instantiate `RunRequest(partition_key=...)` instead.",
        )

        if not (self.partitions_def and self.partitioned_config):
            check.failed("Called run_request_for_partition on a non-partitioned job")

        if (
            isinstance(self.partitions_def, DynamicPartitionsDefinition)
            and self.partitions_def.name
        ):
            # Do not support using run_request_for_partition with dynamic partitions,
            # since this requires querying the instance once per run request for the
            # existent dynamic partitions
            check.failed(
                "run_request_for_partition is not supported for dynamic partitions. Instead, use"
                " RunRequest(partition_key=...)"
            )

        self.partitions_def.validate_partition_key(partition_key, current_time=current_time)

        run_config = (
            run_config
            if run_config is not None
            else self.partitioned_config.get_run_config_for_partition_key(partition_key)
        )
        run_request_tags = {
            **(tags or {}),
            **self.partitioned_config.get_tags_for_partition_key(
                partition_key,
                job_name=self.name,
            ),
        }

        return RunRequest(
            run_key=run_key,
            run_config=run_config,
            tags=run_request_tags,
            job_name=self.name,
            asset_selection=asset_selection,
            partition_key=partition_key,
        )

    def get_config_schema_snapshot(self) -> "ConfigSchemaSnapshot":
        return self.get_job_snapshot().config_schema_snapshot

    def get_job_snapshot(self) -> "JobSnapshot":
        return self.get_job_index().job_snapshot

    def get_job_index(self) -> "JobIndex":
        from dagster._core.host_representation import JobIndex
        from dagster._core.snap import JobSnapshot

        return JobIndex(JobSnapshot.from_job_def(self), self.get_parent_job_snapshot())

    def get_job_snapshot_id(self) -> str:
        return self.get_job_index().job_snapshot_id

    def get_parent_job_snapshot(self) -> Optional["JobSnapshot"]:
        if self.op_selection_data:
            return self.op_selection_data.parent_job_def.get_job_snapshot()
        elif self.asset_selection_data:
            return self.asset_selection_data.parent_job_def.get_job_snapshot()
        else:
            return None

    def has_direct_input_value(self, input_name: str) -> bool:
        return input_name in self.input_values

    def get_direct_input_value(self, input_name: str) -> object:
        if input_name not in self.input_values:
            raise DagsterInvalidInvocationError(
                f"On job '{self.name}', attempted to retrieve input value for input named"
                f" '{input_name}', but no value was provided. Provided input values:"
                f" {sorted(list(self.input_values.keys()))}"
            )
        return self.input_values[input_name]

    def _copy(self, **kwargs: Any) -> "JobDefinition":
        # dict() calls copy dict props
        base_kwargs = dict(
            graph_def=self.graph,
            resource_defs=dict(self.resource_defs),
            executor_def=self._executor_def,
            logger_defs=self._loggers,
            config=self._original_config_argument,
            name=self._name,
            description=self.description,
            tags=self.tags,
            metadata=self._metadata,
            hook_defs=self.hook_defs,
            op_retry_policy=self._op_retry_policy,
            version_strategy=self.version_strategy,
            _subset_selection_data=self._subset_selection_data,
            asset_layer=self.asset_layer,
            input_values=self.input_values,
            partitions_def=self.partitions_def,
            _was_explicitly_provided_resources=None,
        )
        resolved_kwargs = {**base_kwargs, **kwargs}  # base kwargs overwritten for conflicts
        job_def = JobDefinition.dagster_internal_init(**resolved_kwargs)
        update_wrapper(job_def, self, updated=())
        return job_def

    @public
    def with_top_level_resources(
        self, resource_defs: Mapping[str, ResourceDefinition]
    ) -> "JobDefinition":
        """Apply a set of resources to all op instances within the job."""
        resource_defs = check.mapping_param(resource_defs, "resource_defs", key_type=str)
        return self._copy(resource_defs=resource_defs)

    @public
    def with_hooks(self, hook_defs: AbstractSet[HookDefinition]) -> "JobDefinition":
        """Apply a set of hooks to all op instances within the job."""
        hook_defs = check.set_param(hook_defs, "hook_defs", of_type=HookDefinition)
        return self._copy(hook_defs=(hook_defs | self.hook_defs))

    def with_executor_def(self, executor_def: ExecutorDefinition) -> "JobDefinition":
        return self._copy(executor_def=executor_def)

    def with_logger_defs(self, logger_defs: Mapping[str, LoggerDefinition]) -> "JobDefinition":
        return self._copy(logger_defs=logger_defs)

    @property
    def op_selection(self) -> Optional[AbstractSet[str]]:
        return set(self.op_selection_data.op_selection) if self.op_selection_data else None

    @property
    def asset_selection(self) -> Optional[AbstractSet[AssetKey]]:
        return self.asset_selection_data.asset_selection if self.asset_selection_data else None

    @property
    def resolved_op_selection(self) -> Optional[AbstractSet[str]]:
        return self.op_selection_data.resolved_op_selection if self.op_selection_data else None


def _swap_default_io_man(resources: Mapping[str, ResourceDefinition], job: JobDefinition):
    """Used to create the user facing experience of the default io_manager
    switching to in-memory when using execute_in_process.
    """
    from dagster._core.storage.mem_io_manager import mem_io_manager

    if (
        resources.get(DEFAULT_IO_MANAGER_KEY) in [default_job_io_manager]
        and job.version_strategy is None
    ):
        updated_resources = dict(resources)
        updated_resources[DEFAULT_IO_MANAGER_KEY] = mem_io_manager
        return updated_resources

    return resources


@dagster_maintained_io_manager
@io_manager(
    description="Built-in filesystem IO manager that stores and retrieves values using pickling."
)
def default_job_io_manager(init_context: "InitResourceContext"):
    # support overriding the default io manager via environment variables
    module_name = os.getenv("DAGSTER_DEFAULT_IO_MANAGER_MODULE")
    attribute_name = os.getenv("DAGSTER_DEFAULT_IO_MANAGER_ATTRIBUTE")
    silence_failures = os.getenv("DAGSTER_DEFAULT_IO_MANAGER_SILENCE_FAILURES")

    if module_name and attribute_name:
        from dagster._core.execution.build_resources import build_resources

        try:
            module = importlib.import_module(module_name)
            attr = getattr(module, attribute_name)
            check.invariant(
                isinstance(attr, IOManagerDefinition),
                (
                    "DAGSTER_DEFAULT_IO_MANAGER_MODULE and DAGSTER_DEFAULT_IO_MANAGER_ATTRIBUTE"
                    " must specify an IOManagerDefinition"
                ),
            )
            with build_resources({"io_manager": attr}, instance=init_context.instance) as resources:
                return resources.io_manager
        except Exception as e:
            if not silence_failures:
                raise
            else:
                warnings.warn(
                    f"Failed to load io manager override with module: {module_name} attribute:"
                    f" {attribute_name}: {e}\nFalling back to default io manager."
                )

    # normally, default to the fs_io_manager
    from dagster._core.storage.fs_io_manager import PickledObjectFilesystemIOManager

    instance = check.not_none(init_context.instance)
    return PickledObjectFilesystemIOManager(base_dir=instance.storage_directory())


@dagster_maintained_io_manager
@io_manager(
    description="Built-in filesystem IO manager that stores and retrieves values using pickling.",
    config_schema={"base_dir": Field(StringSource, is_required=False)},
)
def default_job_io_manager_with_fs_io_manager_schema(init_context: "InitResourceContext"):
    # support overriding the default io manager via environment variables
    module_name = os.getenv("DAGSTER_DEFAULT_IO_MANAGER_MODULE")
    attribute_name = os.getenv("DAGSTER_DEFAULT_IO_MANAGER_ATTRIBUTE")
    silence_failures = os.getenv("DAGSTER_DEFAULT_IO_MANAGER_SILENCE_FAILURES")

    if module_name and attribute_name:
        from dagster._core.execution.build_resources import build_resources

        try:
            module = importlib.import_module(module_name)
            attr = getattr(module, attribute_name)
            check.invariant(
                isinstance(attr, IOManagerDefinition),
                (
                    "DAGSTER_DEFAULT_IO_MANAGER_MODULE and DAGSTER_DEFAULT_IO_MANAGER_ATTRIBUTE"
                    " must specify an IOManagerDefinition"
                ),
            )
            with build_resources({"io_manager": attr}, instance=init_context.instance) as resources:
                return resources.io_manager
        except Exception as e:
            if not silence_failures:
                raise
            else:
                warnings.warn(
                    f"Failed to load io manager override with module: {module_name} attribute:"
                    f" {attribute_name}: {e}\nFalling back to default io manager."
                )
    from dagster._core.storage.fs_io_manager import PickledObjectFilesystemIOManager

    # normally, default to the fs_io_manager
    base_dir = init_context.resource_config.get(
        "base_dir", init_context.instance.storage_directory() if init_context.instance else None
    )

    return PickledObjectFilesystemIOManager(base_dir=base_dir)


def _config_mapping_with_default_value(
    inner_schema: ConfigType,
    default_config: Mapping[str, Any],
    job_name: str,
) -> ConfigMapping:
    if not isinstance(inner_schema, Shape):
        check.failed("Only Shape (dictionary) config_schema allowed on Job ConfigMapping")

    def config_fn(x):
        return x

    updated_fields = {}
    field_aliases = inner_schema.field_aliases
    for name, field in inner_schema.fields.items():
        if name in default_config:
            updated_fields[name] = Field(
                config=field.config_type,
                default_value=default_config[name],
                description=field.description,
            )
        elif name in field_aliases and field_aliases[name] in default_config:
            updated_fields[name] = Field(
                config=field.config_type,
                default_value=default_config[field_aliases[name]],
                description=field.description,
            )
        else:
            updated_fields[name] = field

    config_schema = Shape(
        fields=updated_fields,
        description=(
            "This run config schema was automatically populated with default values "
            "from `default_config`."
        ),
        field_aliases=inner_schema.field_aliases,
    )

    config_evr = validate_config(config_schema, default_config)
    if not config_evr.success:
        raise DagsterInvalidConfigError(
            f"Error in config when building job '{job_name}' ",
            config_evr.errors,
            default_config,
        )

    return ConfigMapping(
        config_fn=config_fn, config_schema=config_schema, receive_processed_config_values=False
    )


def get_run_config_schema_for_job(
    graph_def: GraphDefinition,
    resource_defs: Mapping[str, ResourceDefinition],
    executor_def: "ExecutorDefinition",
    logger_defs: Mapping[str, LoggerDefinition],
    asset_layer: Optional[AssetLayer],
    was_explicitly_provided_resources: bool = False,
) -> ConfigType:
    return JobDefinition(
        name=graph_def.name,
        graph_def=graph_def,
        resource_defs=resource_defs,
        executor_def=executor_def,
        logger_defs=logger_defs,
        asset_layer=asset_layer,
        _was_explicitly_provided_resources=was_explicitly_provided_resources,
    ).run_config_schema.run_config_schema_type


def _infer_asset_layer_from_source_asset_deps(job_graph_def: GraphDefinition) -> AssetLayer:
    """For non-asset jobs that have some inputs that are fed from SourceAssets, constructs an
    AssetLayer that includes those SourceAssets.
    """
    asset_keys_by_node_input_handle: Dict[NodeInputHandle, AssetKey] = {}
    source_assets_list = []
    source_asset_keys_set = set()
    io_manager_keys_by_asset_key: Mapping[AssetKey, str] = {}

    # each entry is a graph definition and its handle relative to the job root
    stack: List[Tuple[GraphDefinition, Optional[NodeHandle]]] = [(job_graph_def, None)]

    while stack:
        graph_def, parent_node_handle = stack.pop()

        for node_name, input_source_assets in graph_def.node_input_source_assets.items():
            node_handle = NodeHandle(node_name, parent_node_handle)
            for input_name, source_asset in input_source_assets.items():
                if source_asset.key not in source_asset_keys_set:
                    source_asset_keys_set.add(source_asset.key)
                    source_assets_list.append(source_asset)

                input_handle = NodeInputHandle(node_handle, input_name)
                asset_keys_by_node_input_handle[input_handle] = source_asset.key
                for resolved_input_handle in graph_def.node_dict[
                    node_name
                ].definition.resolve_input_to_destinations(input_handle):
                    asset_keys_by_node_input_handle[resolved_input_handle] = source_asset.key

                if source_asset.io_manager_key:
                    io_manager_keys_by_asset_key[source_asset.key] = source_asset.io_manager_key

        for node_name, node in graph_def.node_dict.items():
            if isinstance(node.definition, GraphDefinition):
                stack.append((node.definition, NodeHandle(node_name, parent_node_handle)))

    return AssetLayer(
        asset_keys_by_node_input_handle=asset_keys_by_node_input_handle,
        asset_info_by_node_output_handle={},
        asset_deps={},
        dependency_node_handles_by_asset_key={},
        assets_defs=[],
        source_asset_defs=source_assets_list,
        io_manager_keys_by_asset_key=io_manager_keys_by_asset_key,
        node_output_handles_to_dep_asset_keys={},
        partition_mappings_by_asset_dep={},
    )


def _build_all_node_defs(node_defs: Sequence[NodeDefinition]) -> Mapping[str, NodeDefinition]:
    all_defs: Dict[str, NodeDefinition] = {}
    for current_level_node_def in node_defs:
        for node_def in current_level_node_def.iterate_node_defs():
            if node_def.name in all_defs:
                if all_defs[node_def.name] != node_def:
                    raise DagsterInvalidDefinitionError(
                        'Detected conflicting node definitions with the same name "{name}"'.format(
                            name=node_def.name
                        )
                    )
            else:
                all_defs[node_def.name] = node_def

    return all_defs


def _create_run_config_schema(
    job_def: JobDefinition,
    required_resources: AbstractSet[str],
) -> "RunConfigSchema":
    from .run_config import (
        RunConfigSchemaCreationData,
        construct_config_type_dictionary,
        define_run_config_schema_type,
    )
    from .run_config_schema import RunConfigSchema

    # When executing with a subset job, include the missing nodes
    # from the original job as ignored to allow execution with
    # run config that is valid for the original
    ignored_nodes: Sequence[Node] = []
    if job_def.is_subset:
        if isinstance(job_def.graph, SubselectedGraphDefinition):  # op selection provided
            ignored_nodes = job_def.graph.get_top_level_omitted_nodes()
        elif job_def.asset_selection_data:
            parent_job = job_def
            while parent_job.asset_selection_data:
                parent_job = parent_job.asset_selection_data.parent_job_def

            ignored_nodes = [
                node for node in parent_job.graph.nodes if not job_def.has_node_named(node.name)
            ]
    else:
        ignored_nodes = []

    run_config_schema_type = define_run_config_schema_type(
        RunConfigSchemaCreationData(
            job_name=job_def.name,
            nodes=job_def.graph.nodes,
            graph_def=job_def.graph,
            dependency_structure=job_def.graph.dependency_structure,
            executor_def=job_def.executor_def,
            resource_defs=job_def.resource_defs,
            logger_defs=job_def.loggers,
            ignored_nodes=ignored_nodes,
            required_resources=required_resources,
            direct_inputs=job_def.input_values,
            asset_layer=job_def.asset_layer,
        )
    )

    if job_def.config_mapping:
        outer_config_type = job_def.config_mapping.config_schema.config_type
    else:
        outer_config_type = run_config_schema_type

    if outer_config_type is None:
        check.failed("Unexpected outer_config_type value of None")

    config_type_dict_by_name, config_type_dict_by_key = construct_config_type_dictionary(
        job_def.all_node_defs,
        outer_config_type,
    )

    return RunConfigSchema(
        run_config_schema_type=run_config_schema_type,
        config_type_dict_by_name=config_type_dict_by_name,
        config_type_dict_by_key=config_type_dict_by_key,
        config_mapping=job_def.config_mapping,
    )
